# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the C implementation of the WCA potential.

This test is comparing the C implementation with the PyRETIS
implementation.
"""
# pylint: disable=invalid-name
import unittest
import itertools
import numpy as np
from pyretis.core import Particles, create_box, System
from pyretis.core.units import create_conversion_factors
from pyretis.forcefield import ForceField
from pyretis.forcefield.potentials import DoubleWellWCA
from pyretis.forcefield.potentials import PairLennardJonesCutnp
from pyretis.tools import generate_lattice
from wcafunctions import WCAPotential


def set_up_initial_state():
    """Create particles for the test."""
    create_conversion_factors('lj')
    lattice, size = generate_lattice('sq', [3, 3], density=0.6)
    lattice += np.random.random((lattice.shape)) * 0.05
    size = np.array(size)
    box = create_box(low=size[:, 0], high=size[:, 1], periodic=[True, True])
    system = System(temperature=1.0, units='lj', box=box)
    system.particles = Particles(dim=2)
    for i, pos in enumerate(lattice):
        if i < 2:
            system.add_particle(name='Ar', pos=pos, mass=1.0, ptype=1)
        else:
            system.add_particle(name='Ar', pos=pos, mass=1.0, ptype=0)
    msg = 'Created lattice with {} atoms.'
    print(msg.format(system.particles.npart))
    return system


def set_up_python_forcefield():
    """Create internal Python WCA force field."""
    wca_pot = PairLennardJonesCutnp(dim=2, shift=True, mixing='geometric')
    wca_params = {0: {'sigma': 1.0, 'epsilon': 1.0, 'factor': 2.**(1./6.)},
                  1: {'sigma': 1.0, 'epsilon': 1.0, 'factor': 2.**(1./6.)}}
    dwca_pot = DoubleWellWCA(dim=2)
    dwca_params = {'types': [(1, 1)],
                   'rzero': 1.0 * (2.0**(1.0/6.0)),
                   'height': 15.0, 'width': 0.5}
    forcefield = ForceField('Double well force field, PyRETIS',
                            potential=[wca_pot, dwca_pot],
                            params=[wca_params, dwca_params])
    return forcefield


def set_up_c_forcefield():
    """Create external C WCA force field."""
    wca_pot_c = WCAPotential()
    wca_paramsc = {'sigma': 1.0, 'epsilon': 1.0, 'rcut': 2.**(1./6.),
                   'idxi': 0, 'idxj': 1, 'rzero': 2.0**(1./6),
                   'height': 15.0, 'width': 0.5}
    forcefield = ForceField('Double well force field, PyRETIS + c',
                            potential=[wca_pot_c],
                            params=[wca_paramsc])
    return forcefield


class WCATest(unittest.TestCase):
    """Run the tests for the C potential class."""

    def test_wca_pot(self):
        """Test evaluation of potential."""
        system = set_up_initial_state()
        forcefield_python = set_up_python_forcefield()
        forcefield_c = set_up_c_forcefield()

        pot, force_virial, force_and_pot = [], [], []
        for ffield in [forcefield_python, forcefield_c]:
            system.forcefield = ffield
            vpot1 = system.potential()
            force1, virial1 = system.force()
            vpot2, force2, virial2 = system.potential_and_force()

            self.assertAlmostEqual(vpot1, vpot2)

            forceok = np.allclose(force1, force2)
            self.assertTrue(forceok)

            virialok = np.allclose(virial1, virial2)
            self.assertTrue(virialok)

            pot.append(vpot1)
            force_virial.append((force1, virial1))
            force_and_pot.append((vpot2, force2, virial2))
        for pair in itertools.combinations(range(len(pot)), 2):
            i, j = pair
            self.assertAlmostEqual(pot[i], pot[j])
            virialok = np.allclose(force_virial[i][1], force_virial[j][1])
            self.assertTrue(virialok)
            forceok = np.allclose(force_virial[i][0], force_virial[j][0])
            self.assertTrue(forceok)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""
Example of running a MD NVE simulation.

In this example, we are running backward and forward using
the internal engines/force fields.
"""

# pylint: disable=invalid-name
import sys
from matplotlib import pyplot as plt
from matplotlib import gridspec
from pyretis.core.units import create_conversion_factors
from pyretis.setup import create_simulation
from pyretis.inout.fileio import FileIO
from pyretis.inout.formats import ThermoTableFormatter
# for plotting:
from pyretis.inout.plotting import mpl_set_style
# simulation settings:
settings = {}
settings['simulation'] = {
    'task': 'md-nve',
    'steps': 30,
}
settings['system'] = {
    'units': 'lj',
    'temperature': 2.5,
    'dimensions': 2
}
settings['engine'] = {
    'class': 'velocityverlet',
    'timestep': 0.002
}
settings['output'] = {
    'backup': 'overwrite',
    'energy-file': 1,
    'screen': 10,
    'trajectory-file': 1
}
settings['potential'] = [
    {'class': 'PairLennardJonesCutnp',
     'parameter': {0: {'sigma': 1, 'epsilon': 1, 'factor': 2.5},
                   1: {'sigma': 1, 'epsilon': 1, 'factor': 2.5}},
     'shift': True, 'dim': 2}
]
settings['particles'] = {
    'position': {'input_file': 'initial.gro'},
    'velocity': {'generate': 'maxwell', 'momentum': True, 'seed': 0}
}

create_conversion_factors(settings['system']['units'])
simulation_nve = create_simulation(settings)

# Set up extra output:
thermo_file = FileIO('thermo.txt', 'w', ThermoTableFormatter(), backup=False)
thermo_file.open()
store_results = []
# Also create some other outputs:
simulation_nve.set_up_output(settings, progress=False)
# Run the simulation forward:
for result in simulation_nve.run():
    stepno = result['cycle']['stepno']
    thermo_file.output(stepno, result['thermo'])
    result['thermo']['stepno'] = stepno
    store_results.append(result['thermo'])
# Run backward:
simulation_nve.ensemble['system'].particles.vel *= -1.0
simulation_nve.extend_cycles(settings['simulation']['steps'] - 1)
for result in simulation_nve.run():
    stepno = result['cycle']['stepno']
    thermo_file.output(stepno, result['thermo'])
    result['thermo']['stepno'] = stepno
    store_results.append(result['thermo'])
thermo_file.close()

mpl_set_style()  # Load PyRETIS plotting style
step = [res['stepno'] for res in store_results]
pot_e = [res['vpot'] for res in store_results]
kin_e = [res['ekin'] for res in store_results]
tot_e = [res['etot'] for res in store_results]
pressure = [res['press'] for res in store_results]
temp = [res['temp'] for res in store_results]
# Plot some energies:
fig1 = plt.figure()
gs = gridspec.GridSpec(2, 2)
ax1 = fig1.add_subplot(gs[:, 0])
ax1.plot(step, pot_e, label='Potential')
ax1.plot(step, kin_e, label='Kinetic')
ax1.plot(step, tot_e, label='Total')
ax1.set_xlabel('Step no.')
ax1.set_ylabel('Energy per particle')
ax1.legend(loc='center left', prop={'size': 'small'})
ax2 = fig1.add_subplot(gs[0, 1])
ax2.plot(step, temp)
ax2.set_ylabel('Temperature')
ax3 = fig1.add_subplot(gs[1, 1])
ax3.plot(step, pressure)
ax3.set_xlabel('Step no.')
ax3.set_ylabel('Pressure')
fig1.tight_layout()
if 'noplot' not in sys.argv[1:]:
    plt.show()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains common functions for the visualization.

It contains some functions that are used to plot regression lines
and interface planes, and generate surface plots.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

gen_surface (:py:func:`.gen_surface`)
    Generates a user-defined surface/contour/etc plot with colorbar in
    given matplotlib.figure and -.axes objects.

plot_int_plane(:py:func:`.plot_regline`)
    Generates interface planes for the current span of x-values, in a
    given matplotlib.axes-object.

plot_regline (:py:func:`.plot_regline`)
    Calculates the linear regression and correlation, plots a line for the
    regression in the given matplotlib.axes-object, with info in legend.

_grid_it_up (:py:func:`._grid_it_up`)
    Maps the x,y and z data to a numpy.meshgrid using scipy interpolation
    at a user defined resolution.
"""

import numpy as np
import matplotlib as mpl  # pylint: disable=import-error
from scipy.interpolate import griddata as scgriddata
from scipy.stats import linregress as linreg


def plot_regline(ax, x, y):
    """Plot a regression line calculated from input data in the input subplot.

    Parameters
    ----------
    ax : Matplotlib subplot, where reg.line is to be plotted.
    x, y : list
        Floats, coordinates of data regression lines
        are calculated from.

    Updates
    -------
    Regression line with values.

    """
    xplot = np.linspace(min(x, default=0), max(x, default=0), 2)
    slope, intercept, r_value, _, _ = linreg(x, y)
    rtxt = 'y={0:.2f}x + {1:.2f}, $r^2$={2:.3f}'
    rline = ax.plot(xplot, slope*xplot + intercept, '-', c='black',
                    label=rtxt.format(slope, intercept, r_value**2))
    return rline


def plot_int_plane(ax, pos, yminmax, zminmax, visible=False):
    """Generate the interface planes for 3D visualization.

    Parameters
    ----------
    ax : The matplotlib.axes object where the planes will be plotted.
    pos : float
        The x-axis position of the interface plane.
    yminmax, zminmax : list of float
        The limits of the plane in the 3D canvas.
    visible : boolean, optional
        If True, shows interface planes.

    Returns
    -------
    plane : A 3D surface at x=pos, perpendicular to the x-axis.

    """
    y, z = np.linspace(*yminmax, 2), np.linspace(*zminmax, 2)
    y, z = np.meshgrid(y, z)
    point = np.array([pos, 0.0, 0.0])
    normal = np.array([1.0, 0.0, 0.0])
    dot = -point.dot(normal)
    x = (-normal[2]*y - normal[1]*z - dot) * 1./normal[0]
    plane = ax.plot_surface(x, y, z, color='grey',
                            alpha=0.30, visible=visible)
    return plane


def gen_surface(x, y, z, fig, ax, cbar_ax=None, dim=3, method='contour',
                res_x=400, res_y=400, colormap='viridis'):
    """Generate the chosen surface/contour/scatter plot.

    Parameters
    ----------
    x, y, z : list
        Coordinates of data points. (x,y) the chosen orderP pairs, and
        z is the chosen energy value of the two combinations.
    fig : Matplotlib object
        main canvas.
    ax : Matplotlib object
        axes of the plot.
    cbar_ax : Matplotlib object, optional
        plot color-bar.
    dim: int
        dimension of the plot surface.
    method : string, optional
        Method used for plotting data, default is contour lines.
    res_x, res_y : integer, optional
        Resolution of plot, either as N*N bins in 2D histogram
        (Density plot) or as grid-points for interpolation of data
        (Surface and contour plots).
    colormap : string, optional
        Name of the colormap/color scheme to use when plotting.

    Returns
    -------
    surf : Matplotlib object
         The chosen surface/contour/plot object.
    cbar : Matplotlib object
         The chosen color-bar.

    """
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    # pylint: disable=too-many-locals
    xmin, xmax = min(x, default=0), max(x, default=0)
    ymin, ymax = min(y, default=0), max(y, default=0)
    zmin, zmax = min(z, default=0), max(z, default=0)

    cmap = mpl.colormaps[colormap]

    colors = [cmap(z[i]) for i in range(len(z))]
    if not zmin == zmax:
        colors = [cmap((z[i]-zmin)/(zmax-zmin)) for i in range(len(z))]

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    if dim == 3:
        # 3d plot settings
        ax.set_zlim3d(zmin, zmax)
        ax.zaxis.set_ticklabels([])

    # Methods for plotting in 3D
    if method == 'scatter':
        if dim == 3:
            surf = ax.scatter(x, y, z, c=colors, s=res_x / 100.0)
        else:
            surf = ax.scatter(x, y, c=colors, s=res_x / 100.0)

        norm = mpl.colors.Normalize(vmin=zmin, vmax=zmax)
        cbar = fig.colorbar(
            mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
            cax=cbar_ax)
        return surf, cbar

    g_x, g_y, g_z = _grid_it_up([x, y, z], res_x=res_x, res_y=res_y)
    if method == 'surface':
        surf = ax.plot_surface(g_x, g_y, g_z, cmap=cmap, vmin=zmin, vmax=zmax,
                               facecolor=colors, shade=True,
                               antialiased=False)

    elif method == 'contourf':
        surf = ax.contourf(g_x, g_y, g_z, cmap=cmap)

    else:  # contour
        surf = ax.contour(g_x, g_y, g_z, cmap=cmap)

    cbar = fig.colorbar(surf, cax=cbar_ax)

    return surf, cbar


def _grid_it_up(xyz, res_x=200, res_y=200, fill='max'):
    """Map x, y and z data values to a numpy meshgrid by interpolation.

    Parameters
    ----------
    xyz : list of list
        Lists of data values.
    res_x, res_y : integer, optional
        Resolution (number of points in a axis range).
    fill : string, optional
        Criteria to color the un-explored regions.

    Returns
    -------
    g_x, g_y, g_z : list
        Numpy.arrays of mapped data.

    """
    # Convert 3 columns of data to grid for matplotlib"""
    x_i = np.linspace(min(xyz[0], default=0), max(xyz[0], default=0), res_x)
    y_i = np.linspace(min(xyz[1], default=0), max(xyz[1], default=0), res_y)
    g_x, g_y = np.meshgrid(x_i, y_i)

    fill_value = 0
    # Scipy griddata """ # Works
    if fill == 'max':
        fill_value = max(xyz[2], default=0)
    elif fill == 'min':
        fill_value = min(xyz[2], default=0)
    g_z = scgriddata((xyz[0], xyz[1]), np.array(xyz[2]), (g_x, g_y),
                     method='linear', fill_value=fill_value)

    return g_x, g_y, g_z



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Tools which can help with setting up simulations.

This package defines some simple tools which may be useful for
creating simulations.

Package structure
-----------------

Modules
~~~~~~~

lattice.py (:py:mod:`pyretis.tools.lattice`)
    Defines tools for setting up and generating lattice points.

recalculate_order.py (:py:mod:`pyretis.tools.recalculate_order`)
    Methods for recalculating order parameters on external paths.

Important methods defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

generate_lattice (:py:func:`.generate_lattice`)
    Generates points on a lattice.

recalculate_order (:py:func:`.recalculate_order`)
    Recalculate order parameter(s).
"""
from pyretis.tools.lattice import generate_lattice
from pyretis.tools.recalculate_order import recalculate_order



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains examples for order parameters.

This file is distributed as part of the documentation of PyRETIS.
"""
import numpy as np
from pyretis.orderparameter import OrderParameter


class PlaneDistanceX(OrderParameter):
    """A positional order parameter.

    This class defines a very simple order parameter which is
    the distance from a plane for a given particle.
    """

    def __init__(self, index, plane_position):
        """Initialise the order parameter.

        Parameters
        ----------
        index : integer
            Selects the particle to use.

        plane_position : float
            The location of the plane, along the x-axis.
        """
        txt = 'Distance from particle {} to the plane at {}'.format(
            index,
            plane_position)
        super().__init__(description=txt)
        self.index = index
        self.plane_position = plane_position

    def calculate(self, system):
        """Calculate the order parameter."""
        pos = system.particles.pos[self.index]
        dist = -np.abs(pos[0] - self.plane_position)
        return [dist]



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module defines the class for interfacing LAMMPS.

Important classes defined here
------------------------------

LAMMPSEngine (:py:class:`.LAMMPSEngine`)
    The class responsible for interfacing with LAMMPS.

"""
import logging
import os
import shlex
import numpy as np
from pyretis.core.common import counter
from pyretis.engines.external import ExternalMDEngine
from pyretis.inout.fileio import FileIO
from pyretis.inout.settings import look_for_input_files
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


# Let us define the LAMMPS commands PyRETIS will make use of:

# 1) The commands related to the order parameter.
# This first command is to ensure that the order parameter is
# calculated also for cases where we do not run any MD steps. This
# can for instance be when we are generating velocities:
ORDER_HACK = """run 0 every 1 "print '{}' file {} screen no" # :-)"""
# The parameter is here the variables to print.
# The next command is outputting the order parameter:
ORDER_FIX = 'fix order_output all print {} "{}" append {} screen no'
# The parameters are: frequency, variables to print and the file name.

# 2) The commands related to thermodynamic output:
# Define a thermo style for LAMMPS:
THERMO_STYLE = 'thermo_style custom step temp press pe ke etotal'

# 3) The commands related to trajectory output and intput:
TRAJ_DUMP = 'dump {} all custom {} {} id {}'
# The three parameters are here: dump-id, frequency, output file name
# and the format for what to print. What to print depends on the number
# dimensions we consider and this is defined below:
TRAJ_OUT = {
    1: 'x vx ix',
    2: 'x y vx vy ix iy',
    3: 'x y z vx vy vz ix iy iz',
}
# Modify the format for the output by increasing the number of digits
# used for positions and velocities:
TRAJ_FMT = 'dump_modify {} format {} %23.16g'
# The parameter is here the dump-id and the column to apply it to.
# Add a command to undump the trajectory
TRAJ_UNDUMP = 'undump traj_pyretis'
# We also need to read in trajectories and/or snapshots:
READ_DUMP = 'read_dump {} {} {} box yes'
# The parameters are: the filename and the index in that file.
# The next command will read in LAMMPS restart files:
READ_RESTART = 'read_restart {}'
# The parameter is here the file name.

# 4) The commands for generating velocities:
GENERATE_VEL = 'velocity all create ${SET_TEMP} ${VEL_SEED} dist gaussian'
# The parameters are here the temperature and the random seed, which
# are actually set elsewhere in the LAMMPS input script.
# The next command sets the seed to use:
GENERATE_VEL_SEED = 'variable VEL_SEED equal {}'
# The parameter is here the seed to use.

# 5) The commands for stopping a simulation based on the order parameter
# and the given interface positions:
INTERFACE_RIGHT_FIX = 'fix op_stop_right all halt {} v_op_1 > {}'
# The parameters are here the frequency and the right interface.
INTERFACE_LEFT_FIX = 'fix op_stop_left all halt {} v_op_1 < {}'
# The parameters are here the frequency and the left interface.

# 6) The commands for rescaling energies if this is requested:
RESCALE_ENERGY = [
    'variable ke_rescale equal "v_SET_ENERGY - pe"',
    'variable alpha equal "sqrt(v_ke_rescale / ke)"',
]
# Here we need different commands, depending on the number
# of dimensions we are simulating. These are hard-coded below:
RESCALE_DIM = {
    3: [
        'variable vx atom "v_alpha * vx"',
        'variable vy atom "v_alpha * vy"',
        'variable vz atom "v_alpha * vz"',
        'velocity all set v_vx v_vy v_vz',
        ],
    2: [
        'variable vx atom "v_alpha * vx"',
        'variable vy atom "v_alpha * vy"',
        'velocity all set v_vx v_vy 0.0',
        ],
    1: [
        'variable vx atom "v_alpha * vx"',
        'velocity all set v_vx 0.0 0.0',
        ],
}

# 7) Commands for reversing velocities:
# First the command for reversing:
VEL_REV = 'variable        v{0} atom -v{0}'
# The argument is here to dimension to apply this to, e.g. "x".
# Then, the command to actually set the velocities.
VEL_SET = 'velocity        all set {} {} {}'
# The arguments are the velocities to set, either v_vx, v_vy, v_vz
# or "NULL".


def read_lammps_log(filename):
    """Read some info from a LAMMPS log file.

    In particular, this method is used to read the thermodynamic
    output from a simulation (e.g. potential and kinetic energies).

    Parameters
    ----------
    filename : string
        The path to the LAMMPS log file.

    Returns
    -------
    out : dict
        A dict containing the data we found in the file.

    """
    energy_keys = []
    energy_data = {}
    read_energy = False
    with open(filename, 'r', encoding='utf-8') as logfile:
        for lines in logfile:
            if lines.startswith('Step'):
                # Assume that this is the start of the thermo output.
                energy_keys = [i.strip() for i in lines.strip().split()]
                for key in energy_keys:
                    # Note: This will discard the previously read
                    # thermodynamic data. This is because we only want
                    # to read the final section of thermodynamic data,
                    # which, by construction of the LAMMPS input file,
                    # correspond to the output of the MD run.
                    energy_data[key] = []
                read_energy = True
                continue
            if lines.startswith('Loop time'):
                # Assume this marks the end of the thermo output.
                read_energy = False
                energy_keys = []
                continue
            if read_energy and energy_keys:
                # Assume that we are reading energies.
                try:
                    data = [float(i) for i in lines.strip().split()]
                except ValueError:
                    # Some text has snuck into the thermo output,
                    # ignore it:
                    continue
                for key, value in zip(energy_keys, data):
                    if key == 'Step':
                        energy_data[key].append(int(value))
                    else:
                        energy_data[key].append(value)
    for key, val in energy_data.items():
        energy_data[key] = np.array(val)
    out = {'energy': energy_data}
    return out


def read_lammps_input(filename):
    """Read a LAMMPS input file.

    This will read in a LAMMPS input file and can be used to read
    the value of particular settings, for instance the time step.

    Parameters
    ----------
    filename : string
        The path to the LAMMPS input file.

    Returns
    -------
    out : list of tuples
        The settings found in the LAMMPS input file. The tuples are
        on the form `(keyword, setting)`.

    """
    settings = []
    with open(filename, 'r', encoding='utf-8') as infile:
        for lines in infile:
            current_line, _, _ = lines.strip().partition('#')
            split = current_line.split()
            if split:
                keyword, setting = split[0], split[1:]
                settings.append((keyword, ' '.join(setting)))
    return settings


def add_to_lammps_input(infile, outfile, to_add):
    """Add PyRETIS specific settings to an input file for LAMMPS.

    This method will append settings needed by PyRETIS to an input
    file for LAMMPS.

    Parameters
    ----------
    infile : string
        Path to a file which contains the LAMMPS settings.
    outfile : string
        Path to the file we should create.
    to_add : list of strings
        The settings to add for PyRETIS.

    """
    if os.path.isfile(outfile):
        logger.debug(
            '"%s" exists and will be overwritten by PyRETIS.', outfile
        )
    with open(infile, 'r', encoding='utf-8') as indata:
        data = indata.read()
        with open(outfile, 'w', encoding='utf-8') as output:
            output.write(data)
            output.write('\n'.join(to_add))


def _add_order_fix(settings):
    """Add the fix for printing the order parameter for LAMMPS."""
    lines = []
    if 'order-fix' in settings:
        logger.debug('Adding fix for order parameter to LAMMPS input.')
        lines.append('# Order parameter fix:')
        for line in settings['order-fix']:
            lines.append(line)
    return lines


def _add_traj_dump(settings):
    """Add the dump commands for storing the LAMMPS trajectory."""
    lines = [
        '# PyRETIS trajectory settings:',
        '# PyRETIS requested the following trajectory output:',
    ]
    lines.append(
        TRAJ_DUMP.format(
            'traj_pyretis',
            settings['subcycles'],
            settings['traj'],
            TRAJ_OUT[settings['dimension']],
        )
    )
    # Format for positions:
    idx = 0
    for i in range(int(settings['dimension'])):
        idx = i + 2
        lines.append(TRAJ_FMT.format('traj_pyretis', idx))
    # Format for velocities:
    for i in range(int(settings['dimension'])):
        lines.append(TRAJ_FMT.format('traj_pyretis', idx + 1 + i))
    return lines


def _add_stopping_condition(settings):
    """Add the interface stopping condition to the LAMMPS input."""
    lines = []
    if 'interfaces' in settings:
        lines.append('# Stopping condition fix:')
        lines.append(
            INTERFACE_RIGHT_FIX.format(
                settings['subcycles'],
                settings['interfaces'][-1],
            )
        )
        if abs(settings['interfaces'][0]) != float('inf'):
            lines.append(
                INTERFACE_LEFT_FIX.format(
                    settings['subcycles'],
                    settings['interfaces'][0],
                )
            )
    return lines


def _add_generate_vel(settings):
    """Add generation of velocities to the LAMMPS input."""
    lines = []
    if 'generate_vel' in settings:
        seed = settings['generate_vel'].get('seed', 1)
        lines.append('# PyRETIS requested random velocities')
        lines.append(GENERATE_VEL_SEED.format(seed))
        lines.append(GENERATE_VEL)
        lines.append('run 0')
        if settings['generate_vel'].get('rescale', None):
            for i in RESCALE_ENERGY:
                lines.append(i)
            for i in RESCALE_DIM[settings['dimension']]:
                lines.append(i)
    return lines


def create_lammps_md_input(system, infile, outfile, settings):
    """Create MD input file for LAMMPS.

    We will here write to a new file, by appending text to the
    given template.

    Parameters
    ----------
    system : object like :py:class:`.System`
        The system contains the current particle state, and
        this determines the initial configuration and if we
        are to modify velocities in some way (e.g. reversing
        them before running or drawing new ones).
    infile : string
        Path to a file which contains the LAMMPS settings.
    settings : dict
        The settings we are going to use for creating the
        LAMMPS file.

    Returns
    -------
    out : list of strings
        The LAMMPS commands written to the output file.

    """
    new_lines = ['', '# Added by PyRETIS:', '']
    # Add reading of initial configuration:
    config = system_to_lammps(
        system,
        settings['reverse_velocities'],
        settings['dimension']
    )
    if config:
        new_lines += config
    # Update the thermo output from LAMMPS:
    new_lines.append('# PyRETIS requested the following thermo output:')
    new_lines.append(THERMO_STYLE)
    new_lines.append(f'thermo {settings["subcycles"]}')
    # Add randomization of velocities, if requested:
    new_lines += _add_generate_vel(settings)
    # Add storing of the LAMMPS trajectory:
    new_lines += _add_traj_dump(settings)
    # Add order parameter calculation:
    new_lines += _add_order_fix(settings)
    # Add stopping condition(s):
    new_lines += _add_stopping_condition(settings)
    # Add the number of steps to run:
    new_lines.append('# Requested steps by PyRETIS:')
    new_lines.append(f'run {settings["steps_subcycles"]}')
    # Add "clean-up" to the LAMMPS file:
    new_lines.append('# PyRETIS clean up:')
    new_lines.append(TRAJ_UNDUMP)
    add_to_lammps_input(infile, outfile, new_lines)
    return new_lines


def system_to_lammps(system, reverse_velocities, dimension):
    """Convert a LAMMPS system into LAMMPS commands for loading it.

    This method will convert a system created by LAMMPS into
    LAMMPS commands, so that LAMMPS can read the given snapshot
    and use it.

    Parameters
    ----------
    system : object like :py:class:`.System`
        The system we are converting.
    reverse_velocities : boolean
        True if the velocities in the system are to be reversed.
    dimension : integer
        The number of dimensions used in the LAMMPS simulation.

    Returns
    -------
    out : list of strings
        The commands needed by LAMMPS to read the configuration.

    """
    config = system.particles.get_pos()
    lammps = []
    if config[0].endswith('.restart') and config[1] == 0:
        # Assume that this is a LAMMPS restart file.
        lammps.append('# PyRETIS requested a LAMMPS restart file:')
        lammps.append(READ_RESTART.format(config[0]))
    elif config[0].endswith('.data') and config[1] == 0:
        # Assume that this is a LAMMPS data file AND that there
        # is already a line in the LAMMPS input file reading this
        # file. The motivation behind this is that we define this
        # file type to be the initial configuration and that data files
        # will never be used for something else.
        lammps.append(f'# PyRETIS requested the LAMMPS data file: {config[0]}')
    else:
        # Assume that this is a LAMMPS trajectory.
        lammps.append('# PyRETIS requested the following snapshot:')
        lammps.append(
            READ_DUMP.format(config[0], config[1], TRAJ_OUT[dimension])
        )
        lammps.append('# Reset time step:')
        lammps.append('reset_timestep 0')
    if reverse_velocities:
        # We add commands for reversing the velocities:
        lammps.append('# PyRETIS requested reversing of velocities:')
        vel = ['NULL', 'NULL', 'NULL']
        for i, dim in enumerate(['x', 'y', 'z'][:dimension]):
            lammps.append(VEL_REV.format(dim))
            vel[i] = f'v_v{dim}'
        lammps.append(VEL_SET.format(*vel))
        lammps.append('run 0')
    return lammps


class LAMMPSEngine(ExternalMDEngine):
    """
    A class for interfacing LAMMPS.

    Attributes
    ----------
    lmp : string
        The command for executing LAMMPS
    input_path : string
        The directory where the input files are stored.
    input_files : dict of strings
        The names of the input files.

    """

    needs_order = False

    def __init__(self, lmp, input_path, subcycles, extra_files=None):
        """Set up the LAMMPS engine.

        Parameters
        ----------
        lmp : string
            The LAMMPS executable.
        input_path : string
            The absolute path to where the input files are stored.
        subcycles : integer
            The frequency of output of data by LAMMPS.
        extra_files : list of strings, optional
            Additional files needed to run the LAMMPS simulation.

        """
        # Note: We set time step to zero for now. This will be updated
        # to the correct value when the timestep has been read from
        # the input script:
        super().__init__('LAMMPS Engine', 0.0, subcycles)
        self.lmp = lmp
        self.input_path = os.path.abspath(input_path)
        # Define the files we require:
        input_files = {
            'conf': 'system.data',
            'template': 'lammps.in',  # The template file for LAMMPS.
            'order': 'order.in',  # Definition of the order parameter.
        }
        # The user may choose to structure the LAMMPS input into
        # several files. This is something we will not try to figure
        # out, and we assume that the user knows what he/she is doing.
        # We assume that these extra files will also be present and
        # stored in the given input path. Now, we look for such files:
        self.input_files = look_for_input_files(
            self.input_path,
            input_files,
        )
        self.run_files = {
            'conf': self.input_files['conf'],
            'order': self.input_files['order'],
        }
        if extra_files is not None:
            extra = look_for_input_files(
                self.input_path,
                {f'file-{i}': val for i, val in enumerate(extra_files)}
            )
            for key, val in extra.items():
                self.run_files[key] = val
        # Read the LAMMPS input file:
        self.settings = {
            'dimension': 3,
            'timestep': 0.0
        }
        for (key, val) in read_lammps_input(self.input_files['template']):
            self.settings[key] = val
        # Also set the timestep explicitly in case it is needed:
        self.timestep = self.settings['timestep']

    def _make_order_fix(self, ordername):
        """Create a LAMMPS fix for the order parameter.

        Note that we here make LAMMPS also output for step zero.

        """
        order_settings = read_lammps_input(
            os.path.join(self.input_path, self.input_files['order'])
        )
        ops = ['$(step)']
        for key, val in order_settings:
            if key == 'variable':
                ops.append(f'${{{val.strip().split()[0]}}}')
        txt = [
            ORDER_HACK.format(' '.join(ops), ordername),
            ORDER_FIX.format(self.subcycles, ' '.join(ops), ordername),
        ]
        return txt

    def add_input_files(self, dirname):
        """Add required input files to a given directory.

        Parameters
        ----------
        dirname : string
            The path to the directory where we want to add the files.

        """
        for _, filepathi in self.run_files.items():
            basename = os.path.basename(filepathi)
            dest = os.path.join(dirname, basename)
            if not os.path.isfile(dest):
                logger.debug('Adding input file "%s" to "%s"',
                             basename, dirname)
                self._copyfile(filepathi, dest)

    def read_order_parameters(self, ordername):
        """Read order parameters as calculated by LAMMPS.

        We assume here that these can be found in the current execute
        directory in a file named `ordername`, and further that they
        contain the step number in the first column, followed by the
        order parameters in following columns.

        """
        orderfile = os.path.join(self.exe_dir, ordername)
        order = np.loadtxt(orderfile, ndmin=2)
        # Removing the first column, as this is not an order parameter,
        # but the step number in the simulation. The usage of deleting
        # along axis=1 is the motivation for forcing ndim=2 above.
        order = np.delete(order, 0, 1)
        return order

    def read_energies(self, name):
        """Read energies obtained in a LAMMPS run.

        Here, we assume that the energies can be found in a log file
        in the current execute directory.

        """
        logfile = os.path.join(self.exe_dir, f'{name}.log')
        data = read_lammps_log(logfile)
        return data['energy']

    def _propagate_from(self, name, path, ensemble, msg_file,
                        reverse=False):  # pragma: no cover
        """
        Propagate with LAMMPS from the current system configuration.

        Parameters
        ----------
        name : string
            A name to use for the trajectory we are generating.
        path : object like :py:class:`.PathBase`
            This is the path we use to fill in phase-space points.
        ensemble: dict
            It contains:

            * `system`: object like :py:class:`.System`
              The system object gives the initial state.
            * `order_function`: object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.
            * `interfaces`: list of floats
              These interfaces define the stopping criterion.

        msg_file : object like :py:class:`.FileIO`
            An object we use for writing out messages that are useful
            for inspecting the status of the current propagation.
        reverse : boolean, optional
            If True, the system will be propagated backward in time.

        Returns
        -------
        success : boolean
            This is True if we generated an acceptable path.
        status : string
            A text description of the current status of the propagation.

        """
        return

    def step(self, system, name):  # pragma: no cover
        """Perform a single step with LAMMPS.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we are integrating.
        name : string
            To name the output files from the LAMMPS step.

        Returns
        -------
        out : string
            The name of the output configuration, obtained after
            completing the step.

        """
        return

    def _extract_frame(self, traj_file, idx, out_file):  # pragma: no cover
        return

    @staticmethod
    def _read_configuration(filename):  # pragma: no cover
        return

    def _reverse_velocities(self, filename, outfile):  # pragma: no cover
        """Reverse the velocities for a given configuration."""
        return

    def modify_velocities(self, ensemble, vel_settings):
        """Modify velocities."""
        dek = None
        system = ensemble['system']
        kin_old = system.particles.ekin
        rgen = ensemble['rgen']
        kin_new = None
        if vel_settings.get('aimless', False):
            logger.debug('Modifying velocities with LAMMPS.')
            name = 'generate_vel'
            if rgen:
                seed = rgen.random_integers(1, 2147483647 - 2)
                # Note: `random_integers` is inclusive for both
                # numbers. We don't like to have > 2**31 - 1 as a
                # possibility. For some reason, LAMMPS might also
                # die/hang for a seed equal to 2**31 - 1, so this is
                # why 2 is subtracted here, resulting in the range:
                # [1, 2147483646]
            else:
                seed = 1
            settings = {
                'steps_subcycles': 0,
                'reverse_velocities': False,
                'generate_vel': {
                    'seed': seed,
                    'rescale': vel_settings.get('rescale_energy',
                                                vel_settings.get('rescale'))
                },
            }
            self.run_lammps(system, settings, name)
            kin_new = system.particles.ekin
            if kin_old is None:
                dek = float('inf')
            else:
                dek = kin_new - kin_old
            return dek, kin_new
        raise ValueError(
            'LAMMPS only support the aimless velocity modification.'
        )

    def run_lammps(self, system, settings, name):
        """Execute LAMMPS.

        This method will handle input files, run LAMMPS and
        return some data after the run.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system defines the initial state we are using.
        settings : dict
            This dict contains settings for creating the LAMMPS
            input file.
        name : string
            This string is used as a base name for some of the
            LAMMPS input scripts and output files.

        Returns
        -------
        out[0] : numpy.array
            The order parameters obtained during the run.
        out[1] : dict of numpy.arrays
            The energies, each dictionary key corresponds to a
            specific energy term.
        out[2] : string
            The name of the trajectory created by LAMMPS in this run.

        """
        logger.debug('Adding input files for LAMMPS.')
        self.add_input_files(self.exe_dir)
        # Name the trajectory output:
        settings['traj'] = os.path.join(self.exe_dir, f'{name}.lammpstrj')
        settings['order'] = f'order_{name}.txt'
        # Add the subcycles & timestep settings:
        settings['subcycles'] = self.subcycles
        settings['timestep'] = self.timestep
        settings['dimension'] = int(self.settings['dimension'])
        # Add the order-fix:
        settings['order-fix'] = self._make_order_fix(settings['order'])
        # Name the input script for LAMMPS:
        input_file = f'{name}.in'
        script_file = os.path.join(self.exe_dir, input_file)
        # Create the input file we will use or running LAMMPS:
        create_lammps_md_input(
            system,
            self.input_files['template'],
            script_file,
            settings,
        )
        cmd = shlex.split(self.lmp)
        cmd_arguments = [
            '-in', input_file,
            '-l', f'{name}.log',
            '-screen', f'{name}.screen'
        ]
        cmd.extend(cmd_arguments)
        logger.debug('Executing LAMMPS "%s".', name)
        self.execute_command(cmd, cwd=self.exe_dir)
        logger.debug('Reading LAMMPS order parameters & energies after run.')
        order = self.read_order_parameters(settings['order'])
        energy = self.read_energies(name)
        # Set the state of the system to the last point:
        system.order = order[-1]
        system.particles.ekin = energy['KinEng'][-1]
        system.particles.vpot = energy['PotEng'][-1]
        system.particles.set_pos(
            (settings['traj'], settings['steps_subcycles'])
        )
        return order, energy, settings['traj']

    def integrate(self, system, steps):  # order_function=None, thermo='full'):
        """Propagate several integration steps with LAMMPS."""
        logger.debug('Integrating with LAMMPS.')
        # Add settings for this run:
        settings = {
            'steps_subcycles': steps * self.subcycles,
            'reverse_velocities': system.particles.get_vel(),
        }
        # Execute LAMMPS:
        self.run_lammps(system, settings, 'pyretis_md')

    def propagate(self, path, ensemble, reverse=False):
        """
        Propagate the equations of motion with the external code.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            This is the path we use to fill in phase-space points.
            We are here not returning a new path - this since we want
            to delegate the creation of the path to the method
            that is running `propagate`.
        ensemble: dict
            It contains:

            * `system`: object like :py:class:`.System`
              The system object gives the initial state for the
              integration. The initial state is stored and the system is
              reset to the initial state when the integration is done.
            * `order_function`: object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.
            * `interfaces`: list of floats
              These interfaces define the stopping criterion.

        reverse : boolean, optional
            If True, the system will be propagated backward in time.

        Returns
        -------
        success : boolean
            This is True if we generated an acceptable path.
        status : string
            A text description of the current status of the propagation.

        """
        initial_state = ensemble['system']
        interfaces = ensemble['interfaces']
        logger.debug('Running propagate with: "%s"', self.description)
        prefix = str(counter())
        if reverse:
            logger.debug('Running backward in time.')
            name = prefix + '_trajB'
        else:
            logger.debug('Running forward in time.')
            name = prefix + '_trajF'
        logger.debug('Trajectory name: "%s"', name)
        # Also create a message file for inspecting progress:
        msg_file_name = os.path.join(self.exe_dir, f'msg-{name}.txt')
        logger.debug('Writing propagation progress to: %s', msg_file_name)
        msg_file = FileIO(msg_file_name, 'w', None, backup=False)
        msg_file.open()
        msg_file.write(f'# Preparing propagation with {self.description}')
        msg_file.write(f'# Trajectory label: {name}')

        system = initial_state.copy()
        logger.debug('Initial state: %s', system)
        # For storing LAMMPS settings:
        settings = {
            'steps_subcycles': path.maxlen * self.subcycles,
            'interfaces': interfaces,
            'reverse_velocities': reverse != system.particles.vel_rev,
        }

        if settings['reverse_velocities']:
            logger.debug('Reversing velocities in initial config.')
        system.particles.set_vel(reverse)

        # Propagate from the current point:
        msg_file.write(f'# Interfaces: {interfaces}')
        msg_file.write(f'# Running LAMMPS in folder: {self.exe_dir}')
        order, energy, traj = self.run_lammps(system, settings, name)
        msg_file.write('# Done running LAMMPS, adding points to path.')
        for i, orderi in enumerate(order):
            phase_point = system.copy()
            phase_point.order = orderi
            phase_point.particles.ekin = energy['KinEng'][i]
            phase_point.particles.vpot = energy['PotEng'][i]
            phase_point.particles.set_pos((traj, i * self.subcycles))
            phase_point.particles.set_vel(reverse)
            status, success, stop, _ = self.add_to_path(
                path,
                phase_point,
                interfaces[0],
                interfaces[-1]
            )
            if stop:
                break
        msg_file.write('# Propagation with LAMMPS all done.')
        msg_file.close()
        return success, status

    def dump_phasepoint(self, phasepoint, deffnm='conf'):
        """Dump a phase point to a new file.

        Note
        ----
        We do not reverse the velocities here.

        """
        settings = {'steps_subcycles': 0, 'reverse_velocities': False}
        self.run_lammps(phasepoint, settings, deffnm)

    def calculate_order(self, ensemble):
        """Return the last seen order parameter."""
        return ensemble['system'].order



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This package contains functions for input/output for the analysis.

The functions defined here will typically run the analysis on the
given input and write outputs, typically this will be plots and simple
text files.

Package structure
-----------------

Modules
~~~~~~~

analysisio.py (:py:mod:`.analysisio`)
    Methods that will output results from the analysis functions.
    The methods defined here can also be used to run an analysis on
    output files from PyRETIS.

__init__.py
    This file handles imports for PyRETIS.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

analyse_file (:py:func:`.analyse_file`)
    Method to analyse a file.

run_analysis_files (:py:func:`.run_analysis_files`)
    Method to analyse simulation data in output files. This is a
    post-processing step.
"""
from .analysisio import analyse_file, run_analysis_files



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Some common functions for generating simple tables and numbers.

This module contains some common functions for the generation of
reports. The functions defined here are typically used to format
numbers and generate tables for the reports.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

generate_rst_table (:py:func:`.generate_rst_table`)
    Generate reStructuredText for a table.

generate_latex_table (:py:func:`.generate_latex_table`)
    Generate latex code for a table.

latexify_number (:py:func:`.latexify_number`)
    Change exponential notation into something nicer for latex.

"""


__all__ = ['generate_rst_table', 'generate_latex_table', 'latexify_number']


def generate_rst_table(table, title, headings):
    """Generate reStructuredText for a table.

    This is a general function to generate a table in reStructuredText.
    The table is specified with a title, headings for the columns and
    the contents of the columns and rows.

    Parameters
    ----------
    table : list of lists
        `table[i][j]` is the contents of column `j` of row `i` of the
        table.
    title : string
        The header/title for the table.
    headings : list of strings
        These are the headings for each table column.

    """
    # for each column, we need to figure out how wide it should be
    # 1) Check headings
    col_len = [len(col) + 2 for col in headings]
    # 2) Check if some of the columns are wider
    for row in table:
        for i, col in enumerate(row):
            if len(col) >= col_len[i] - 2:
                col_len[i] = len(col) + 2  # add some extra space

    width = len(col_len) + sum(col_len) - 1
    topline = '+' + width * '-' + '+'
    # create the header
    str_header = f'|{{0:<{width}s}}|'
    str_header = str_header.format(title)
    # make format for header:
    head_fmt = [f'{{0:^{col}s}}' for col in col_len]
    # create first row, which is the header on columns:
    row_line = [fmt.format(col) for fmt, col in zip(head_fmt, headings)]
    row_line = [''] + row_line + ['']
    row_line = '|'.join(row_line)
    # also set-up the horizontal line:
    hline = [''] + [col * '-' for col in col_len] + ['']
    hline = '+'.join(hline)
    # generate table
    str_table = [topline, str_header, hline, row_line, hline.replace('-', '=')]
    for row in table:
        row_line = [fmt.format(col) for fmt, col in zip(head_fmt, row)]
        row_line = [''] + row_line + ['']
        row_line = '|'.join(row_line)
        str_table.extend([row_line, hline])
    return str_table


def generate_latex_table(table, title, headings, fixnum=None):
    r"""Generate latex code for a table.

    This function will generate latex code for a table. The table is
    given with a title, headings for the columns and the contents of
    the table. For latex we might wish to make some numbers more pretty
    by removing exponential notation: i.e. ``1.e-10`` can be replaced
    by ``1.0 \times 10^{-10}`` (which should render like
    :math:`1.0 \times 10^{-10}`).

    Parameters
    ----------
    table : list of lists
        `table[i][j]` is the contents of column `j` of row `i` of the
        table.
    title : string
        The header/title for the table.
    headings : list of strings
        These are the headings for each table column.
    fixnum : list of integers, optional
        These integers identify the columns where `latexify_number` is
        to be applied.

    """
    str_table = [
        r'\begin{longtable}{' + len(headings) * '| c ' + '|}',
        r'\caption{' + title + r'}\\',
        r'\hline',
        r' & '.join(headings) + r'\\ \hline',
    ]
    for row in table:
        rowl = row
        if fixnum:
            rowl = [latexify_number(col) if i in fixnum else col for i, col
                    in enumerate(row)]
        str_table.append(' & '.join(rowl) + r'\\')
    str_table.append(r'\hline')
    str_table.append(r'\end{longtable}')
    return str_table


def latexify_number(str_float):
    r"""Change exponential notation into something nicer for latex.

    This will change exponential notation, e.g ``1.2e-03``, into
    ``1.2 \times 10^{-3}`` for latex output which should be rendered
    like :math:`1.2 \times 10^{-3}`.

    Parameters
    ----------
    str_float : string
        This is the string representation of a float.

    Returns
    -------
    out : string
        A formatted string for latex.

    """
    if 'e' in str_float:
        base, exp = str_float.split('e')
        return fr'${base} \times 10^{{{int(exp)}}}$'
    return fr'${str_float}$'



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two pandas dataframe files."""
import os
import sys


def main():
    """Compare all order.txt and order.txt_000 in each ensemble.

    Returns
    -------
    out : int
        0 if the comparison was successful, 1 otherwise.

    """
    for ensembles in os.listdir('.'):
        if ensembles.isdigit():
            with open(ensembles + '/order.txt') as one:
                with open(ensembles + '/order.txt_000') as two:
                    for line1, line2 in zip(one, two):
                        if line1.split()[1] != line2.split()[1]:
                            if round(float(line1.split()[1]), 2) != \
                                    round(float(line2.split()[1]), 2):
                                print('File differs')
                                return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Method for generating reports from path sampling simulations.

The reports are useful for displaying results from the analysis.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

generate_report_tis (:py:func:`.generate_report_tis`)
    Generate a report for the overall results from a TIS simulation.

generate_report_tis_path (:py:func:`.generate_report_tis_path`)
    Generate a report for a single TIS simulation.

generate_report_retis (:py:func:`.generate_report_retis`)
    Generate a report for the overall results from a RETIS simulation.
"""
import logging
from pyretis.inout.report.markup import (generate_rst_table,
                                         generate_latex_table)
from pyretis.inout.formats.formatter import apply_format, format_number
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['generate_report_tis', 'generate_report_tis_path',
           'generate_report_retis']


def generate_report_tis_path(analysis, output='rst'):
    """Generate a report for a single TIS simulation.

    Parameters
    ----------
    analysis : dict
        This is the output from the analysis.
    output : string, optional
        This is the desired output format. It must match one of the
        formats defined in
        :py:const:`pyretis.inout.report.report._TEMPLATES`. Default
        is 'rst' (reStructuredText).

    Returns
    -------
    out[0] : string
        The generated report in the desired format.
    out[1] : string
        The file extension (i.e. file type) for the generated report.

    """
    result = analysis['pathensemble']
    report = {'ensemble': result['out']['ensemble'],
              'figures': _get_path_figures(result['figures']),
              'tables': {'interfaces': None,
                         'probability': None,
                         'path': None,
                         'efficiency': None}}
    # Create tables
    results = [result]
    tables = report['tables']
    tables['interfaces'] = _table_interface(results, fmt=output)[1]
    tables['probability'] = _table_probability(results, fmt=output)[1]
    tables['efficiency'] = _table_efficiencies(results, fmt=output)[1]
    return report


def generate_report_pptis_path(analysis, output='rst'):
    """Generate a report for a single TIS simulation.

    Parameters
    ----------
    analysis : dict
        This is the output from the analysis.
    output : string, optional
        This is the desired output format. It must match one of the
        formats defined in
        :py:const:`pyretis.inout.report.report._TEMPLATES`. Default
        is 'rst' (reStructuredText).

    Returns
    -------
    out[0] : string
        The generated report in the desired format.
    out[1] : string
        The file extension (i.e. file type) for the generated report.

    """
    result = analysis['pathensemble_repptis']
    report = {'ensemble': result['out']['ensemble'],
              'figures': _get_path_figures(result['figures']),
              'tables': {'interfaces': None,
                         'probability': None,
                         'path': None,
                         'efficiency': None}}
    # Create tables
    results = [result]
    tables = report['tables']
    tables['interfaces'] = _table_interface(results, fmt=output)[1]
    tables['probability'] = _table_probability_repptis(results, fmt=output)[1]
    tables['efficiency'] = _table_efficiencies(results, fmt=output)[1]
    return report


def generate_report_tis(analysis, output='rst'):
    """Generate a report for the over-all results from a TIS simulation.

    Parameters
    ----------
    analysis : dict
        This is the output from the analysis.
    output : string, optional
        This is the desired output format. It must match one of the
        formats defined in
        :py:const:`pyretis.inout.report.report._TEMPLATES`. Default
        is 'rst' (reStructuredText).

    Returns
    -------
    out[0] : string
        The generated report in the desired format.
    out[1] : string
        The file extension (i.e. file type) for the generated report.

    """
    report = {'figures': {'tis': [],
                          'matched': None},
              'tables': {'interfaces': None,
                         'probability': None,
                         'path': None,
                         'efficiency': None},
              'numbers': {'pcross': None, 'perr': None}}
    results = analysis['pathensemble']
    figures = report['figures']
    # Add figures:
    for result in results:
        figures['tis'].append(_get_path_figures(result['figures']))
    # Get matched result:
    matched_fig = analysis['matched']['figures']
    matched_out = analysis['matched']['out']
    figures['matched'] = matched_fig.get('matched-probability', None)
    figures['total'] = matched_fig.get('total-probability', None)
    figures['progress'] = matched_fig.get('overall-prun', None)
    figures['error'] = matched_fig.get('overall-err', None)
    # Get numbers:
    report['numbers']['pcross'] = format_number(matched_out['prob'], 0.1, 1)
    report['numbers']['perr'] = format_number(matched_out['relerror'] * 100,
                                              0.1, 100)
    # Get tables:
    tables = report['tables']
    tables['interfaces'] = _table_interface(results, fmt=output)[1]
    tables['probability'] = _table_probability(results, fmt=output)[1]
    tables['efficiency'] = _table_efficiencies(results, fmt=output)[1]
    return report


def generate_report_retis(analysis, output='rst'):
    """Generate a report for the over-all results from a RETIS simulation.

    Parameters
    ----------
    analysis : dict
        This is the output from the analysis.
    output : string, optional
        This is the desired output format. It must match one of the
        formats defined in
        :py:const:`pyretis.inout.report.report._TEMPLATES`. Default
        is 'rst' (reStructuredText).

    Returns
    -------
    out[0] : string
        The generated report in the desired format.
    out[1] : string
        The file extension (i.e. file type) for the generated report.

    """
    report = {'figures': {'tis': [],
                          'matched': None},
              'tables': {'interfaces': None,
                         'interfaces0': None,
                         'path0': None,
                         'probability': None,
                         'path': None,
                         'efficiency': None,
                         'summary': None},
              'numbers': {'pcross': None, 'perr': None, 'flux': None,
                          'flux-err': None},
              'text': {'flux-unit': None}}
    result0 = analysis['pathensemble0']
    results = [result0] + analysis['pathensemble']
    figures = report['figures']
    permeability = result0['out'].get('permeability', False)
    # Add figures:
    for result in results[1:]:
        figures['tis'].append(_get_path_figures(result['figures']))
    figures['path0'] = _get_path_figures(result0['figures'])
    # Get matched result:
    matched_fig = analysis['matched']['figures']
    matched_out = analysis['matched']['out']
    figures['matched'] = matched_fig.get('matched-probability', None)
    figures['total'] = matched_fig.get('total-probability', None)
    figures['progress'] = matched_fig.get('overall-rrun', None)
    figures['error'] = matched_fig.get('overall-err', None)
    if permeability:
        figures['xi'] = analysis['xi']['fig']['xirun']
        figures['xierror'] = analysis['xi']['fig']['xierror']
        figures['tau'] = analysis['tau']['fig']['taurun']
        figures['tauerror'] = analysis['tau']['fig']['tauerror']
        figures['tauref'] = analysis['tau']['fig']['tauref']

    # Get numbers:
    numbers = report['numbers']
    numbers['pcross'] = format_number(matched_out['prob'], 0.1, 1)
    numbers['perr'] = format_number(matched_out['relerror'] * 100, 0.1, 100)
    numbers['flux'] = format_number(analysis['flux']['value'], 0.1, 100)
    numbers['flux-err'] = format_number(analysis['flux']['error'] * 100,
                                        0.1, 100)
    numbers['rate'] = format_number(analysis['rate']['value'], 0.1, 100)
    numbers['rate-err'] = format_number(analysis['rate']['error'] * 100,
                                        0.1, 100)
    numbers['permeability'] = permeability
    if permeability:
        numbers['xi'] = format_number(analysis['xi']['value'], 0.1, 100)
        numbers['xi-err'] = format_number(analysis['xi']['error'] * 100,
                                          0.1, 100)
        numbers['tau'] = format_number(analysis['tau']['value'], 0.1, 100)
        numbers['tau-err'] = format_number(analysis['tau']['error'] * 100,
                                           0.1, 100)
        numbers['perm'] = format_number(analysis['perm']['value'], 0.1, 100)
        numbers['perm-err'] = format_number(analysis['perm']['error'] * 100,
                                            0.1, 100)

    report['text']['flux-unit'] = analysis['flux']['unit']
    # Get tables:
    tables = report['tables']
    tables['interfaces'] = _table_interface(results, fmt=output)[1]
    tables['probability'] = _table_probability(results[1:], fmt=output)[1]
    tables['efficiency'] = _table_efficiencies(results, fmt=output)[1]
    tables['summary'] = _table_summary(report, fmt=output)[1]
    return report


def generate_report_retis0(analysis, output='txt'):
    """Generate a report for [0^-] in a RETIS/REPTIS simulation.

    This method is primarily intended for displaying results on the
    screen during the analysis.

    Parameters
    ----------
    analysis : dict
        This is the output of the analysis.
    output : string, optional
        This is the desired output format. It must match one of the
        formats defined in
        :py:const:`pyretis.inout.report.report._TEMPLATES`. Default
        is 'rst' (reStructuredText).

    Returns
    -------
    out[0] : string
        The generated report in the desired format.
    out[1] : string
        The file extension (i.e. file type) for the generated report.

    """
    if output != 'txt':
        output = 'txt'
        logger.warning('Report for [0^-] is only indended as "txt"'
                       '\nOutput format "%s" ignored', output)
    if 'pathensemble' in analysis:
        result = analysis['pathensemble']
    else:
        result = analysis['pathensemble_repptis']
    flux = result['out']['fluxlength']
    report = {
        'ensemble': result['out']['ensemble'],
        'numbers': {'fluxlength': format_number(flux[0], 0, 10000),
                    'fluxlengtherror': format_number(flux[1] * 100, 0, 100)},
        'tables': {'interfaces0': _table_interface0([result], fmt=output)[1]},
    }
    return report


def generate_report_repptis(analysis, output='rst'):
    """Generate a report for the over-all results from a PPRETIS simulation.

    Parameters
    ----------
    analysis : dict
        This is the output from the analysis.
    output : string, optional
        This is the desired output format. It must match one of the
        formats defined in
        :py:const:`pyretis.inout.report.report._TEMPLATES`. Default
        is 'rst' (reStructuredText).

    Returns
    -------
    out[0] : string
        The generated report in the desired format.
    out[1] : string
        The file extension (i.e. file type) for the generated report.

    """
    report = {'figures': {'tis': [],
                          'matched': None},
              'tables': {'interfaces': None,
                         'probability': None,
                         'path': None,
                         'efficiency': None,
                         'summary': None},
              'numbers': {'pcross': None, 'perr': None, 'flux': None,
                          'flux-err': None},
              'text': {'flux-unit': None}}

    result0 = analysis['pathensemble0']
    results = [result0] + analysis['pathensemble']
    figures = report['figures']
    permeability = result0['out'].get('permeability', False)
    # Add figures:
    for result in results[1:]:
        figures['tis'].append(_get_path_figures(result['figures']))
    figures['path0'] = _get_path_figures(result0['figures'])
    # Get matched result:
    matched_fig = analysis['matched_fig']
    figures['matched'] = matched_fig.get('matched-probability', None)
    figures['progress'] = matched_fig.get('overall-prun', None)
    figures['error'] = matched_fig.get('overall-err', None)
    if permeability:
        figures['xi'] = analysis['xi']['fig']['xirun']
        figures['xierror'] = analysis['xi']['fig']['xierror']
        figures['tau'] = analysis['tau']['fig']['taurun']
        figures['tauerror'] = analysis['tau']['fig']['tauerror']
        figures['tauref'] = analysis['tau']['fig']['tauref']
    # Get numbers:
    numbers = report['numbers']
    numbers['pcross'] = format_number(analysis['cross']['value'], 0.1, 1)
    numbers['perr'] = format_number(analysis['cross']['error'] * 100,
                                    0.1, 100)
    numbers['flux'] = format_number(analysis['flux']['value'], 0.1, 100)
    numbers['flux-err'] = format_number(analysis['flux']['error'] * 100,
                                        0.1, 100)
    numbers['rate'] = format_number(analysis['rate']['value'], 0.1, 100)
    numbers['rate-err'] = format_number(analysis['rate']['error'] * 100,
                                        0.1, 100)
    numbers['permeability'] = permeability
    if permeability:
        numbers['xi'] = format_number(analysis['xi']['value'], 0.1, 100)
        numbers['xi-err'] = format_number(analysis['xi']['error'] * 100,
                                          0.1, 100)
        numbers['tau'] = format_number(analysis['tau']['value'], 0.1, 100)
        numbers['tau-err'] = format_number(analysis['tau']['error'] * 100,
                                           0.1, 100)
        numbers['perm'] = format_number(analysis['perm']['value'], 0.1, 100)
        numbers['perm-err'] = format_number(analysis['perm']['error'] * 100,
                                            0.1, 100)

    report['text']['flux-unit'] = 'time-unit'  # matched['flux']['unit']
    # Get tables:
    tables = report['tables']
    tables['interfaces'] = _table_interface_repptis(results, fmt=output)[1]
    tables['probability'] = _table_probability_repptis(results[1:],
                                                       fmt=output)[1]
    tables['efficiency'] = _table_efficiencies(results, fmt=output)[1]
    tables['summary'] = _table_summary(report, fmt=output)[1]
    return report


def _table_interface_repptis(results, fmt='rst'):
    """Generate the table for the interfaces.

    This table will display the location of the different interfaces.

    Parameters
    ----------
    results : list of dicts
        These are the results of the analysis.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in the desired format which represents
        the table.

    """
    table = []
    for idx, result in enumerate(results):
        interf = result['out']['interfaces']
        # To set the L interface correctly for PPRETIS
        if idx > 1:
            interf0 = results[idx-1]['out']['interfaces'][1]
        else:
            interf0 = interf[0]
        detect = result['out']['detect']
        row = [
            f"{result['out']['ensemble']:^8s}",
            apply_format(interf0, '{:^8.4f}'),
            apply_format(interf[1], '{:^8.4f}'),
        ]
        intf2 = interf[2] if detect is None else detect
        row.append(apply_format(intf2, '{:^8.4f}'))
        table.append(row)
    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, 'Interfaces',
                                         ['Ensemble', 'Left', 'Middle',
                                          'Detect'],
                                         fixnum={0, 1, 2, 3})
    else:
        table_str = generate_rst_table(table, 'Interfaces',
                                       ['Ensemble', 'Left', 'Middle',
                                        'Detect'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _table_interface(results, fmt='rst'):
    """Generate the table for the interfaces.

    This table will display the location of the different interfaces.

    Parameters
    ----------
    results : list of dicts
        These are the results of the analysis.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in the desired format which represents
        the table.

    """
    table = []
    for result in results:
        interf = result['out']['interfaces']
        row = [
            f"{result['out']['ensemble']:^8s}",
            apply_format(interf[0], '{:^8.4f}'),
            apply_format(interf[1], '{:^8.4f}'),
        ]
        detect = result['out']['detect']
        if detect is None:
            row.append('')
        else:
            row.append(apply_format(detect, '{:^8.4f}'))
        table.append(row)
    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, 'Interfaces',
                                         ['Ensemble', 'Left', 'Middle',
                                          'Detect'],
                                         fixnum={0, 1, 2, 3})
    else:
        table_str = generate_rst_table(table, 'Interfaces',
                                       ['Ensemble', 'Left', 'Middle',
                                        'Detect'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _table_interface0(results, fmt='rst'):
    """Generate the table for the [0^-] ensemble.

    This table will display the location of the different interfaces.

    Parameters
    ----------
    results : list of dicts
        These are the results of the analysis.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in the desired format which represents
        the table.

    """
    table = []
    for result in results:
        row = [f"{result['out']['ensemble']:^8s}"]
        interf = result['out']['interfaces']
        row.append(apply_format(interf[0], '{:^8.4f}'))
        row.append(apply_format(interf[1], '{:^8.4f}'))
        row.append(apply_format(interf[2], '{:^8.4f}'))
        table.append(row)
    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, 'Interfaces',
                                         ['Ensemble', 'Left', 'Middle',
                                          'Right'],
                                         fixnum={0, 1, 2, 3})
    else:
        table_str = generate_rst_table(table, 'Interfaces',
                                       ['Ensemble', 'Left', 'Middle',
                                        'Right'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _table_probability(results, fmt='rst'):
    """Generate the table for the probabilities.

    This table will display the crossing probabilities with
    uncertainties.

    Parameters
    ----------
    results : list of dicts
        The dictionaries are the results obtained from the analysis.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in reStructuredText format which represents
        the table.

    """
    table = []
    for result in results:
        row = [
            f"{result['out']['ensemble']:^8s}",
            apply_format(result['out']['prun'][-1], '{:^10.6f}'),
            apply_format(result['out']['blockerror'][2], '{:^10.6f}'),
            apply_format(result['out']['blockerror'][4] * 100, '{:^10.6f}'),
        ]
        table.append(row)
    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, r'Crossing probabilities',
                                         [r'Ensemble', r'$P_\text{cross}$',
                                          r'Error', r'Rel. error (\%)'],
                                         fixnum={0, 1, 2, 3})
    else:
        table_str = generate_rst_table(table, r'Crossing probabilities',
                                       [r'Ensemble', r'Pcross', r'Error',
                                        r'Rel. error (%)'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _table_probability_repptis(results, fmt='rst'):
    """Generate the table for the probabilities.

    This table will display the crossing probabilities with
    uncertainties.

    Parameters
    ----------
    results : list of dicts
        The dictionaries are the results obtained from the analysis.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in reStructuredText format which represents
        the table.

    """
    table = []
    for result in results:
        row = [
            f"{result['out']['ensemble']:^8s}",
            apply_format(result['out']['prun_sl'][-1], '{:^10.6f}'),
            apply_format(result['out']['prun_sr'][-1], '{:^10.6f}'),
            apply_format(result['out']['blockerror_sl'][2], '{:^10.6f}'),
            apply_format(result['out']['blockerror_sr'][2], '{:^10.6f}'),
        ]
        table.append(row)
    if fmt in ['tex', 'latex']:
        str1 = r'$\left\langle p^{ \pm}\right\rangle_{\text {full }}$'
        str2 = r'$\left\langle p^{ \mp}\right\rangle_{\text {full }}$'
        table_str = generate_latex_table(table, r'Probabilities',
                                         [r'Ensemble', str1, str2,
                                          r'Error$^{ \pm}$',
                                          r'Error$^{ \mp}$'],
                                         fixnum={0, 1, 2, 3, 4})
    else:
        table_str = generate_rst_table(table, r'Probabilities',
                                       [r'Ensemble', r'Pcross pm',
                                        r'Pcross mp', r'Error',
                                        r'Rel. error (%)'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _table_efficiencies(results, fmt='rst'):
    """Generate table for efficiencies.

    This table will display results for the efficiencies, acceptance
    ratios and correlation.

    Parameters
    ----------
    results : list of dicts
        The dictionaries are the results obtained from the analysis.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in reStructuredText format which represents
        the table.

    """
    table = []
    for result in results:
        hist1 = result['out']['pathlength'][0]
        row = [
            f"{result['out']['ensemble']:^8s}",
            apply_format(result['out']['tis-cycles'], '{:^10.0f}'),
            apply_format(result['out']['efficiency'][0], '{:^10.6f}'),
            apply_format(result['out']['efficiency'][1], '{:^10.6f}'),
        ]
        row.append(apply_format(hist1[2][0], '{:^10.6f}'))
        table.append(row)
    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, 'Pathensemble data',
                                         ['Ensemble', 'TIS cycles',
                                          'Shoot acc. ratio',
                                          'Swap acc. ratio',
                                          'Avg. path length'],
                                         fixnum={0, 1, 2, 3, 4, 5})
    else:
        table_str = generate_rst_table(table, 'Pathensemble data',
                                       ['Ensemble', 'TIS cycles',
                                        'Shoot acc. ratio',
                                        'Swap acc. ratio',
                                        'Avg. path length'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _table_summary(report, fmt='rst'):
    """Generate table with summary of main results.

    This table will display results for the crossing probability,
    the initial flux and rate constant.

    Parameters
    ----------
    report : dict
        Dict with current report. We use the information in this
        dictionary to generate the table.
    fmt : string, optional
        Determines if we create reStructuredText ('rst') or
        latex ('tex').

    Returns
    -------
    out[0] : list of strings
        These are the rows of the table.
    out[1] : string
        This is a string in reStructuredText format which represents
        the table.

    """
    numbers = report['numbers']
    text = report['text']
    table = [['Crossing probability', numbers['pcross'], numbers['perr']],
             [f"Flux (1/{text['flux-unit']})", numbers['flux'],
              numbers['flux-err']],
             [f"Rate constant (1/{text['flux-unit']})",
              numbers['rate'], numbers['rate-err']]]
    if fmt in ['tex', 'latex']:
        table_str = generate_latex_table(table, 'Summary of main results',
                                         ['Property', 'Value',
                                          r'Relative error ($\%$)'],
                                         fixnum={1, 2})
    else:
        table_str = generate_rst_table(table, 'Summary of main results',
                                       ['Property', 'Value',
                                        'Relative error (%)'])
    table_txt = '\n'.join(table_str)
    return table, table_txt


def _get_path_figures(figures):
    """Return path figures from a dict of figures.

    This method extracts figures from results and makes them available
    to the report.

    Parameters
    ----------
    figures : dict
        The figures generated by the analysis.

    Returns
    -------
    path_figures : dict
        A dict which can be used in the report.

    """
    path_figures = {}
    for fig in {'pcross', 'prun', 'perror', 'perror_sl',
                'perror_sr', 'lpath', 'prun_sl',
                'prun_sr', 'shoots'}:
        for key in figures:
            if key.endswith(fig):
                path_figures[fig] = figures[key]
    return path_figures



# -*- coding: utf-8 -*-
"""Plot results from time tests of the Lennard-Jones potential.

Here we assume that the results are available in files named
``timeings.txt``.
"""
import numpy as np
from matplotlib import pylab as plt


data = {'Serial': np.loadtxt('timings-plain.txt'),
        'OpenMP, 2 threads': np.loadtxt('timings-OMP_NUM_THREADS_2.txt')}

markers = ['o', 's', '^', 'v']

fig = plt.figure(figsize=(12, 5))
ax1 = fig.add_subplot(121)
for i, key in enumerate(data):
    result = data[key]
    npart, avg, std = result[:, 0], result[:, 2], result[:, 3]
    ax1.errorbar(npart, avg, yerr=std, label=key, lw=3, markersize=9,
                 marker=markers[i], alpha=0.9)
ax1.legend(loc='best')
ax1.set_xlabel('System size')
ax1.set_ylabel('Time')

ax2 = fig.add_subplot(122)
norm_avg = data['OpenMP, 2 threads'][:, 2]

for i, key in enumerate(data):
    result = data[key]
    npart, avg = result[:, 0], result[:, 2]
    nmin = min(len(norm_avg), len(avg))
    avg_n = avg[:nmin] / norm_avg[:nmin]
    ax2.plot(npart, avg_n, lw=3, label=key, markersize=9, alpha=0.9,
             marker=markers[i])
ax2.legend(loc='best', bbox_to_anchor=(0.3, 0.7))
ax2.set_xlabel('System size')
ax2.set_ylabel('Time relative to OpenMP with 2 threads')
fig.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the OpenMMEngine class."""
import unittest
import os
import numpy as np
import pyretis
from pyretis.engines import OpenMMEngine
from pyretis.engines.openmm import HAS_OPENMM, make_pyretis_system
from pyretis.core.path import Path
from pyretis.orderparameter import Distance
from pyretis.core.random_gen import RandomGenerator
from pyretis.setup.createsystem import create_system
from .test_engines import (prepare_test_system,
                           MockRandomGenerator)


if HAS_OPENMM:
    from .test_helpers.test_helpers import (create_openmm_simulation,
                                            write_test_pdb, FakeOp,
                                            write_openmm_simulation)


@unittest.skipIf(HAS_OPENMM is False, "OpenMM is not installed")
class TestOpenMMEngine(unittest.TestCase):
    """Test the OpenMMEngine."""
    pdb = 'test_openmm_pyretis'
    f_name = pdb+'.pdb'

    openmm_simulation = 'simulation'
    openmm_module = 'openmm_module.py'

    @classmethod
    def setUpClass(cls):
        if os.path.isfile(cls.f_name):
            i = 1
            while os.path.isfile(cls.pdb+str(i)+'.pdb'):
                i += 1
            cls.f_name = cls.pdb + str(i) + '.pdb'
        write_test_pdb(cls.f_name)
        write_openmm_simulation(cls.openmm_module, cls.f_name)

    @classmethod
    def tearDownClass(cls):
        files = [cls.f_name, cls.openmm_module]
        for f in files:
            if os.path.isfile(f):
                os.remove(f)

    def setUp(self):
        self.sim = create_openmm_simulation(self.f_name)
        # Distance between the two water oxygens periodic= false, because
        # triclininc boxes are hard:
        self.order_parameter = Distance((0, 3), periodic=False)
        self.settings = {'engine': {
            'openmm_module': self.openmm_module,
            'openmm_simulation': self.openmm_simulation}
                         }

        self.system = make_pyretis_system(self.settings)

    def test_init(self):
        """Test initiation of the OpenMMEngine."""
        eng = OpenMMEngine(openmm_simulation=self.sim, subcycles=4)
        assert eng.simulation is self.sim
        assert eng.subcycles == 4

    def test_init_module(self):
        eng = OpenMMEngine(openmm_simulation=self.openmm_simulation,
                           openmm_module=self.openmm_module)
        # Test if it runs
        eng.integration_step(self.system)

    def test_init_string_but_no_module(self):
        with self.assertRaises(RuntimeError):
            OpenMMEngine(openmm_simulation=self.openmm_simulation)

    def test_installation_error(self):
        """Test failure when the OpenMMEngine is missing."""
        with self.assertRaises(RuntimeError):
            pyretis.engines.openmm.HAS_OPENMM = False
            _ = OpenMMEngine(openmm_simulation=self.sim)
        # Reset change:
        pyretis.engines.openmm.HAS_OPENMM = True

    def test_create_system(self):
        settings = self.settings.copy()
        settings['engine']['type'] = 'OpenMM'
        create_system(settings)

    def test_integration_step(self):
        """Test the integration step of the OpenMMEngine."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        eng.integration_step(self.system)
        # Test if resetting works:
        eng.integration_step(self.system)

    def test_integration_step_no_box(self):
        """Test the integration step of the OpenMMEngine with no box."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        self.system.box = None
        eng.integration_step(self.system)
        # Test if resetting works:
        eng.integration_step(self.system)
        assert self.system.box is None

    def test_calculate_order(self):
        """Test order parameter calculation with the OpenMMEngine."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        answer = eng.calculate_order({'order_function': self.order_parameter,
                                      'system': self.system})
        assert self.order_parameter.calculate(self.system) == answer

    def test_calculate_order_xyz(self):
        """Test order parameter calculation with the OpenMMEngine."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        xyz = np.array([[1, 0, 0], [0, 0, 0], [0, 0, 0],
                        [-1, 0, 0], [0, 0, 0], [0, 0, 0]])
        answer = eng.calculate_order({'order_function': self.order_parameter,
                                      'system': self.system},
                                     xyz=xyz,
                                     vel=self.system.particles.vel,
                                     box=self.system.box.length)
        assert answer == [2]

    def test_pass_functions(self):
        """Test "empty" OpenMMEngine methods."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        eng.clean_up()
        eng.dump_phasepoint(1)

    def test_kick_across_middle_from_smaller(self):
        """Test the intialisation method with the OpenMMEngine."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        tis_settings = {'zero_momentum': True,
                        'rescale_energy': False}
        out = eng.kick_across_middle({'system': self.system,
                                      'order_function': self.order_parameter,
                                      'rgen': RandomGenerator()},
                                     middle=3.6,
                                     tis_settings=tis_settings)
        xyz0 = out[0].particles.get_pos()
        xyz1 = out[1].particles.get_pos()
        answer0 = eng.calculate_order({'order_function': self.order_parameter,
                                       'system': self.system},
                                      xyz=xyz0,
                                      vel=self.system.particles.vel,
                                      box=self.system.box.length)[0]
        answer1 = eng.calculate_order({'order_function': self.order_parameter,
                                       'system': self.system},
                                      xyz=xyz1,
                                      vel=self.system.particles.vel,
                                      box=self.system.box.length)[0]
        assert answer0 < answer1
        assert answer0 < 3.6
        assert answer1 > 3.6

    def test_kick_across_middle_from_bigger(self):
        """Test the intialisation method with the OpenMMEngine."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        tis_settings = {'zero_momentum': True,
                        'rescale_energy': False}
        out = eng.kick_across_middle({'system': self.system,
                                      'order_function': self.order_parameter,
                                      'rgen': RandomGenerator()},
                                     middle=3.55,
                                     tis_settings=tis_settings)
        xyz0 = out[0].particles.get_pos()
        xyz1 = out[1].particles.get_pos()
        answer0 = eng.calculate_order({'order_function': self.order_parameter,
                                       'system': self.system},
                                      xyz=xyz0,
                                      vel=self.system.particles.vel,
                                      box=self.system.box.length)[0]
        answer1 = eng.calculate_order({'order_function': self.order_parameter,
                                       'system': self.system},
                                      xyz=xyz1,
                                      vel=self.system.particles.vel,
                                      box=self.system.box.length)[0]
        assert answer0 > answer1
        assert answer0 > 3.55
        assert answer1 < 3.55

    def test_kick_across_middle_if_on_interface(self):
        """Test the intialisation method with the OpenMMEngine."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        tis_settings = {'zero_momentum': True,
                        'rescale_energy': False}

        # Init frame is ON the interface:
        middle = eng.calculate_order({'order_function': self.order_parameter,
                                      'system': self.system},
                                     xyz=self.system.particles.pos,
                                     vel=self.system.particles.vel,
                                     box=self.system.box.length)[0]

        out = eng.kick_across_middle({'system': self.system,
                                      'order_function': self.order_parameter,
                                      'rgen': RandomGenerator()},
                                     middle=middle,
                                     tis_settings=tis_settings)
        xyz0 = out[0].particles.get_pos()
        xyz1 = out[1].particles.get_pos()
        answer0 = eng.calculate_order({'order_function': self.order_parameter,
                                       'system': self.system},
                                      xyz=xyz0,
                                      vel=self.system.particles.vel,
                                      box=self.system.box.length)[0]
        answer1 = eng.calculate_order({'order_function': self.order_parameter,
                                       'system': self.system},
                                      xyz=xyz1,
                                      vel=self.system.particles.vel,
                                      box=self.system.box.length)[0]
        # Order of fluctuation around OP is not set:
        if answer0 < answer1:
            assert answer0 > middle
            assert answer1 < middle
        else:
            assert answer1 < middle
            assert answer0 > middle

    def test_kick_across_middle_if_first_away(self):
        """Test the intialisation method with the OpenMMEngine."""
        eng = OpenMMEngine(openmm_simulation=self.sim)
        tis_settings = {'zero_momentum': True,
                        'rescale_energy': False}

        # 2 op calculations per frame force OP to run away from:
        order_parameter = FakeOp([1, 1, 2, 2, -1, -1])
        _ = eng.kick_across_middle({'system': self.system,
                                    'order_function': order_parameter,
                                    'rgen': RandomGenerator()},
                                   middle=0,
                                   tis_settings=tis_settings)
        # Should have done 5 calls:
        assert order_parameter.n == 5

    def test_modify_velocities(self):
        """Test that we can modify velocities with the OpenMMEngine.

        Copied from pyretis.test.engines.test_engine.py
        """
        system = prepare_test_system()
        eng = OpenMMEngine(1)
        rgen = MockRandomGenerator(seed=1)
        eng.modify_velocities({'system': system, 'rgen': rgen},
                              {'sigma_v': None, 'aimless': True,
                               'momentum': False, 'rescale': None})
        self.assertAlmostEqual(system.particles.vel[0][0], rgen.rgen[1])
        deltav = np.ones_like(system.particles.vel)
        eng.modify_velocities({'system': system, 'rgen': rgen},
                              {'sigma_v': deltav, 'aimless': False,
                               'momentum': False, 'rescale': None})
        self.assertAlmostEqual(system.particles.vel[0][0],
                               rgen.rgen[1] + rgen.rgen[2])
        eng.modify_velocities({'system': system, 'rgen': rgen},
                              {'sigma_v': None, 'aimless': True,
                               'momentum': True, 'rescale': None})
        self.assertAlmostEqual(system.particles.vel[0][0], 0.0)

        system.particles.pos = np.ones_like(system.particles.pos)
        system.particles.vpot = system.potential()
        dek, kin_new = eng.modify_velocities({'system': system, 'rgen': rgen},
                                             {'sigma_v': None, 'aimless': True,
                                              'momentum': False, 'rescale': 6})
        self.assertAlmostEqual(6, kin_new + system.particles.vpot)
        kin_old = kin_new
        dek, kin_new2 = eng.modify_velocities({'system': system, 'rgen': rgen},
                                              {'sigma_v': None,
                                               'aimless': True,
                                               'momentum': False,
                                               'rescale': -1})
        self.assertAlmostEqual(kin_old, kin_new2)
        self.assertAlmostEqual(dek, 0.0)

    def propagate_test(self, reverse=False):
        """Test the propagation method with the OpenMMEngine."""
        path = Path(None, maxlen=10)
        ifaces = [1, 2, 3]
        order_parameter = FakeOp([1, 2, 2.5, 3, 4])
        eng = OpenMMEngine(openmm_simulation=self.sim)
        ensemble = {'system': self.system,
                    'order_function': order_parameter,
                    'interfaces': ifaces}

        eng.propagate(path=path, ensemble=ensemble, reverse=reverse)

        assert order_parameter.n == 5

    def test_propagate_forward(self):
        """Test the propagation method (forward) with the OpenMMEngine."""
        self.propagate_test(reverse=False)

    def test_propagate_backward(self):
        """Test the propagation method (backward) with the OpenMMEngine."""
        self.propagate_test(reverse=True)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of using a integration routine implemented in C."""
import logging
import os
import sys
from pyretis.engines import MDEngine
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
try:
    import vvintegrator
except ImportError:
    MSG = ('Could not import external C library.'
           '\nPlease compile with: "python setup.py build_ext --inplace"')
    logger.critical(MSG)
    raise ImportError(MSG)


__all__ = ['VelocityVerletC']


class VelocityVerletC(MDEngine):
    """VelocityVerletC(MDEngine).

    This class defines the Velocity Verlet integrator.

    Attributes
    ----------
    timestep : float
        The time step.
    half_timestep : float
        The half of the time step.
    desc : string
        Description of the integrator.

    """

    def __init__(self, timestep,
                 desc='The velocity verlet integrator (C)'):
        """Set up the Velocity Verlet integrator.

        Parameters
        ----------
        timestep : float
            The time step.
        desc : string
            Description of the integrator.

        """
        super().__init__(timestep, desc, dynamics='NVE')
        self.half_timestep = self.timestep * 0.5

    def integration_step(self, system):
        """Velocity Verlet integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in `system.particles`.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        particles = system.particles
        vvintegrator.step1(
            particles.pos,
            particles.vel,
            particles.force,
            particles.imass,
            self.timestep,
            self.half_timestep,
            particles.npart,
            particles.dim
        )
        system.potential_and_force()
        vvintegrator.step2(
            particles.vel,
            particles.force,
            particles.imass,
            self.half_timestep,
            particles.npart,
            particles.dim
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file defines the order parameter used for the GROMACS example."""
import logging
import numpy as np
from numpy import average, rint, dot, sqrt
from pyretis.orderparameter import OrderParameter
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class RingDiffusion(OrderParameter):
    """RingDiffusion(OrderParameter).

    This class defines the ring diffusion order parameter for
    the hydrate example.

    Attributes
    ----------
    name : string
        A human-readable name for the order parameter.
    index : integer
        This selects the particle to use for the order parameter.
    periodic : boolean
        This determines if periodic boundaries should be applied to
        the position or not.

    """

    def __init__(self):
        """Initialise the order parameter.

        Parameters
        ----------
        index : tuple of ints
            This is the indices of the atom we will use the position of.
        periodic : boolean, optional
            This determines if periodic boundary conditions should be
            applied to the position.

        """
        super().__init__(description='Ring diffusion for hydrate')
        self.idx1 = np.array([56, 64, 104, 112, 200, 208], dtype='i4')
        # convert to atom index:
        self.idx1 *= 4
        self.idx1 -= 3
        # convert to 0 index:
        self.idx1 -= 1
        self.idx2 = np.array([56, 64, 72, 80, 104, 112, 136, 152, 168, 176,
                              200, 208, 232, 248, 264, 272, 296, 304, 328,
                              336, 344, 352, 360, 368], dtype='i4')
        # convert to atom index:
        self.idx2 *= 4
        self.idx2 -= 3
        # convert to 0 index:
        self.idx2 -= 1
        self.idxd = 1472  # index for diffusing atom

    def calculate(self, system):
        """Calculate the order parameter.

        Here, the order parameter is just the distance between two
        particles.

        Parameters
        ----------
        system : object like `System` from `pyretis.core.system`
            This object is used for the actual calculation, typically
            only `system.particles.pos` and/or `system.particles.vel`
            will be used. In some cases `system.forcefield` can also be
            used to include specific energies for the order parameter.

        Returns
        -------
        out : float
            The order parameter.

        """
        pos = system.particles.pos
        resl = 1.0e3
        cm1 = average(rint(pos[self.idx1] * resl) / resl, axis=0)
        cm2 = average(rint(pos[self.idx2] * resl) / resl, axis=0)
        cmvec = cm2 - cm1
        molvec = rint(pos[self.idxd] * resl) / resl
        molvec -= cm1
        orderp = dot(cmvec, molvec) / sqrt(dot(cmvec, cmvec))
        return [-1.0 * orderp]



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test functionality for the random generator classes."""
import logging
import unittest
import math
import numpy as np
from numpy.random import RandomState
from pyretis.core.system import System
from pyretis.core.box import create_box
from pyretis.core.particles import Particles
from pyretis.core.particlefunctions import (
    calculate_linear_momentum,
    calculate_kinetic_temperature,
)
from pyretis.core.random_gen import (
    RandomGenerator,
    MockRandomGenerator,
    ReservoirSampler,
    RandomGeneratorBorg,
    MockRandomGeneratorBorg,
    create_random_generator,
)
logging.disable(logging.CRITICAL)


class RandomTest(unittest.TestCase):
    """Run the tests for the random classes."""

    def test_rand(self):
        """Test that we can draw random numbers in [0, 1)."""
        rgen = RandomGenerator(seed=0)
        for i in (0, 1, 10, 100, 1000000):
            numbers = rgen.rand(shape=i)
            left = all([j >= 0 for j in numbers])
            self.assertTrue(left)
            right = all([j < 1 for j in numbers])
            self.assertTrue(right)
            self.assertEqual(i, len(numbers))

        # Without arguments
        number = rgen.rand()
        self.assertEqual(1, len(number))
        # Test that it fails as we expect:
        args = [1, 2]
        self.assertRaises(TypeError, rgen.rand, *args)
        self.assertRaises(TypeError, rgen.rand, (1, 2))

    def test_random_integers(self):
        """Test generation for [low, high]."""
        rgen = RandomGenerator(seed=0)
        for i in (-5, 0, 10, 100):
            for j in (-5, 0, 10, 100):
                if i >= j + 1:
                    args = [i, j]
                    self.assertRaises(ValueError, rgen.random_integers,
                                      *args)
                else:
                    for _ in range(100):  # just repeat a bit
                        number = rgen.random_integers(i, j)
                        self.assertTrue(i <= number <= j)
        # Just draw ones:
        numbers = [rgen.random_integers(1, 1) for _ in range(10)]
        self.assertTrue(all([i == 1 for i in numbers]))
        # Just draw 1 or 2
        numbers = [rgen.random_integers(1, 2) for _ in range(100)]
        self.assertTrue(all([i in (1, 2) for i in numbers]))

    def test_random_normal(self):
        """Test generation of numbers from normal distribution."""
        rgen = RandomGenerator(seed=0)
        loc = 1.2345
        std = 0.2468
        numbers = rgen.normal(loc=loc, scale=std, size=100000)
        self.assertAlmostEqual(loc, np.average(numbers), delta=0.01)
        self.assertAlmostEqual(std, np.std(numbers), delta=0.01)

    def test_random_normal_shape(self):
        """Test drawing of numbers from normal distribution with shape."""
        rgen = RandomGenerator(seed=0)
        shape = (10, 3)
        numbers = rgen.normal(loc=0.0, scale=2.0, size=shape)
        self.assertEqual(shape, numbers.shape)

        # Pretend that we have 6 particles with different "mass"
        sigma = [1.0, 2.0, 4.0, 8.0, 16.0, 32.0]
        tol = 0.1
        for dim in (1, 2, 3):
            # lets draw numbers a couple of times:
            pos = []
            for _ in range(1000):
                numbers = rgen.normal(loc=0.0, scale=np.repeat(sigma, dim))
                numbers.shape = (len(sigma), dim)
                pos.append(numbers)
            pos = np.array(pos)
            # std over all drawn matrices:
            std = np.std(pos, axis=(0,))
            # compare for each dimension:
            for i in range(dim):
                std_diff = np.abs(std[:, i] - sigma) / sigma
                self.assertTrue(all([i < tol for i in std_diff]))

    def test_multivariate_normal(self):
        """Just test that we can draw from the multivariate distribution."""
        rgen = RandomGenerator(seed=0)
        mean = np.array([[1.0, 0.0], [0.0, 1.0]])
        cov = np.array([[1.0, 0.0], [0.0, 1.0]])
        numbers = rgen.multivariate_normal(mean, cov)
        self.assertEqual(numbers.shape, (1, 2, 2))
        numbers = rgen.multivariate_normal(mean, cov, size=2)
        self.assertEqual(numbers.shape, (2, 2, 2))

    def test_draw_maxwellian_velocities(self):
        """Test that we can draw with the system object as input."""
        temperature = 2.0
        mass = np.array([1.0, 2.0, 4.0, 16.0, 256.0, 65536.0])
        sigv = np.sqrt(temperature / mass)
        tol = 0.1
        rgen = RandomGenerator(seed=0)
        for dim in (1, 2, 3):
            system = System(
                temperature=temperature,
                units='reduced',
                box=create_box(periodic=[False]*dim))
            system.particles = Particles(dim=dim)
            for i in mass:
                system.add_particle(np.zeros(dim), mass=i, name='Ar', ptype=0)
            all_vel = []
            for _ in range(1000):
                veli, _ = rgen.draw_maxwellian_velocities(system)
                all_vel.append(veli)
            vel = np.array(all_vel)
            # std over all drawn matrices:
            std = np.std(vel, axis=(0,))
            # compare for each dimension:
            for i in range(dim):
                std_diff = np.abs(std[:, i] - sigv) / sigv
                self.assertTrue(all([i < tol for i in std_diff]))

    def test_generate_maxwellian(self):
        """Test the generate_maxwellian_velocities method."""
        particles = Particles(dim=3)
        dof = [1., 1., 1.]
        heavy = []
        light = []
        for i in range(10):
            if i % 2 == 0:
                mass = 1
                light.append(i)
            else:
                mass = 2
                heavy.append(i)
            particles.add_particle(np.zeros(3), np.zeros(3),
                                   np.zeros(3), mass=mass)
        rgen = RandomGenerator(seed=0)
        for mom in (True, False):
            rgen.generate_maxwellian_velocities(particles, 1.0, 2.0,
                                                dof=dof, momentum=mom)
            _, temp, _ = calculate_kinetic_temperature(particles, 1.0,
                                                       dof=dof)
            self.assertAlmostEqual(temp, 2.0)
            close = np.allclose(np.zeros(3),
                                calculate_linear_momentum(particles))
            if mom:
                self.assertTrue(mom)
            else:
                self.assertFalse(mom)
        rgen.generate_maxwellian_velocities(particles, 1.0, 2.0,
                                            dof=dof, momentum=True,
                                            selection=heavy)
        rgen.generate_maxwellian_velocities(particles, 1.0, 2.0,
                                            dof=None, momentum=False,
                                            selection=light)
        _, temp, _ = calculate_kinetic_temperature(particles, 1.0,
                                                   dof=dof)
        self.assertAlmostEqual(temp, 2.0)
        close = np.allclose(
            np.zeros(3),
            calculate_linear_momentum(particles, selection=heavy)
        )
        self.assertTrue(close)
        close = np.allclose(
            np.zeros(3),
            calculate_linear_momentum(particles, selection=light)
        )
        self.assertFalse(close)

    def test_state(self):
        """Test that we can set and get the state of the generator."""
        rgen = RandomGenerator(seed=123)
        for _ in range(5):
            rgen.random_integers(1, 1000)
        state = rgen.get_state()
        numbers1 = [rgen.random_integers(1, 1000) for _ in range(10)]
        rgen.set_state(state)
        numbers2 = [rgen.random_integers(1, 1000) for _ in range(10)]
        for i, j in zip(numbers1, numbers2):
            self.assertEqual(i, j)

    def test_create(self):
        """Test that we create random generators from settings."""
        settings = {}
        rgen = create_random_generator(settings)
        self.assertEqual(rgen.seed, 0)
        self.assertIsInstance(rgen, RandomGenerator)

        settings = {'seed': 100}
        rgen = create_random_generator(settings)
        self.assertEqual(rgen.seed, 100)
        self.assertIsInstance(rgen, RandomGenerator)

        settings = {'rgen': 'mock', 'seed': 101}
        rgen = create_random_generator(settings)
        self.assertEqual(rgen.seed, 101)
        self.assertIsInstance(rgen, MockRandomGenerator)


class TestReservoirSampler(unittest.TestCase):
    """Run the tests for the ReservoirSampler class."""

    def test_init(self):
        """Test the initialisation."""
        rgen = ReservoirSampler()
        self.assertIsInstance(rgen.rgen, RandomState)
        rgen = ReservoirSampler(length=10, rgen=RandomGenerator(seed=1))
        self.assertIsInstance(rgen.rgen, RandomGenerator)

    def test_add_get(self):
        """Test that we can add to the reservoir."""
        rgen = ReservoirSampler(length=10, rgen=MockRandomGenerator(seed=1))
        correct = [21, 0, 0, 0, 2, 2, 18, 29, 1, 0]
        for i in range(30):
            rgen.append(i)
            self.assertEqual(len(rgen.reservoir), 10)
        for i, j in zip(correct, rgen.reservoir):
            self.assertEqual(i, j)
        for i in range(10):
            j = rgen.get_item()
            self.assertEqual(correct[i], j)
        logging.disable(logging.INFO)
        with self.assertLogs('pyretis.core.random_gen', level='CRITICAL'):
            j = rgen.get_item()
        self.assertEqual(correct[0], j)
        logging.disable(logging.CRITICAL)


class TestMockRandomGenerator(unittest.TestCase):
    """Run some tests for the MockRandomGenerator."""

    def test_state(self):
        """Test that we can get and set the state of the generator."""
        rgen = MockRandomGenerator(seed=987)
        for _ in range(5):
            rgen.random_integers(1, 100)
        state = rgen.get_state()
        numbers1 = [rgen.random_integers(1, 100) for _ in range(10)]
        rgen.set_state(state)
        numbers2 = [rgen.random_integers(1, 100) for _ in range(10)]
        for i, j in zip(numbers1, numbers2):
            self.assertEqual(i, j)

    def test_rand(self):
        """Test that we can draw fake random numbers in [0, 1)."""
        rgen = MockRandomGenerator(seed=0)
        numbers = rgen.rand(shape=5)
        self.assertTrue(np.allclose(numbers, rgen.rgen[0:5]))

    def test_random_integers(self):
        """Test that we can draw fake random integers."""
        rgen = MockRandomGenerator(seed=0)
        correct = [14, 10, 14, 15, 13]
        for i in correct:
            j = rgen.random_integers(10, 15)
            self.assertEqual(i, j)
        rgen = MockRandomGenerator(seed=0)
        for _ in range(len(rgen.rgen)):
            j = rgen.random_integers(4, 9)
            self.assertTrue(4 <= j <= 9)

    def test_normal(self):
        """Test that we can draw fake normal numbers."""
        rgen = MockRandomGenerator(seed=0)
        numbers = rgen.normal()
        self.assertAlmostEqual(numbers[0], rgen.rgen[0])
        numbers = rgen.normal(loc=1.0, scale=10, size=5)
        for i, j in zip(numbers, rgen.rgen[1:6]):
            self.assertAlmostEqual(i, j)

    def test_multivariate_normal(self):
        """Test that we can draw fake normal numbers."""
        rgen = MockRandomGenerator(seed=0)
        correct = np.array([[0.0178008, 0.01044599]])
        numbers = rgen.multivariate_normal(1.0, None)
        self.assertTrue(np.allclose(correct, numbers))
        numbers = rgen.multivariate_normal([1.0, 1.0], None, size=3)
        correct = np.array([[0.01765968, 0.01976767],
                            [0.01537996, 0.01986571],
                            [0.01363436, 0.01553565]])
        self.assertTrue(np.allclose(correct, numbers))

    def test_choice(self):
        a = [1, 2, 3, 4]
        rgen = MockRandomGenerator(seed=0)
        correct = [4, 1, 4, 4, 3, 4, 2]
        numbers = rgen.choice(a, 7)
        self.assertTrue(np.allclose(numbers, correct))

    def test_choice_no_size(self):
        a = [1, 2, 3, 4]
        rgen = MockRandomGenerator(seed=0)
        correct = [4]
        numbers = rgen.choice(a)
        self.assertTrue(np.allclose(numbers, correct))

    def test_choice_no_replace(self):
        a = [1, 2, 3, 4]
        rgen = MockRandomGenerator(seed=0)
        correct = [4, 1, 3, 2]
        numbers = rgen.choice(a, size=4, replace=False)
        self.assertTrue(np.allclose(numbers, correct))

    def test_choice_altered_p(self):
        a = [1, 2, 3, 4]
        rgen = MockRandomGenerator(seed=0)
        correct = [1, 1, 1, 1]
        numbers = rgen.choice(a, size=4, p=[1, 0, 0, 0])
        self.assertTrue(np.allclose(numbers, correct))


class TestBorgRandomGenerators(unittest.TestCase):
    """Test the state sharing random generators."""

    def test_borg_mock(self):
        """Test that we share the Mock state."""
        rgens = [
            MockRandomGeneratorBorg(
                seed=i,
                norm_shift=(i == 0),
            ) for i in range(5)
        ]
        state0 = rgens[0].get_state()['state']
        # Are all states the same?
        self.assertTrue(all([i.get_state()['state'] is state0 for i in rgens]))
        # Did the norm_shift get set to the first value?
        self.assertTrue(all([i.norm_shift for i in rgens]))
        # Change the state of one member by requesting a number:
        _ = rgens[0].rand()
        state1 = rgens[0].get_state()['state']
        # Check that the state did change:
        self.assertNotEqual(state0, state1)
        # Are all states still the same?
        self.assertTrue(all([i.get_state()['state'] == state1 for i in rgens]))
        # Set state manually
        rgens[-1].set_state({'state': 3})
        self.assertTrue(all([i.get_state()['state'] == 3 for i in rgens]))
        MockRandomGeneratorBorg.reset_state()

    @staticmethod
    def assert_equal_npstates(state1, state2):
        """Compare two random states coming from numpy.random.get_state()."""
        if not len(state1) == len(state2):
            raise AssertionError('Length of state data differ.')
        if not state1[0] == 'MT19937':
            raise AssertionError('Unexpected state string.')
        if not state1[0] == state2[0]:
            raise AssertionError('States differ for state string.')
        if not np.allclose(state1[1], state2[1]):
            raise AssertionError('States differ for state integers.')
        if not state1[2] == state2[2]:
            raise AssertionError('States differ for "pos" integer.')
        if not state1[3] == state2[3]:
            raise AssertionError('States differ for "has_gauss" integer.')
        if not math.isclose(state1[4], state2[4]):
            raise AssertionError('States differ for "cached_gaussian" float.')

    def test_borg_randomgen(self):
        """Test that we share the Mock state."""
        rgens = [
            RandomGeneratorBorg(seed=i) for i in range(5)
        ]
        # Check that rgen objects are identical:
        obj = rgens[0].rgen
        for i in rgens:
            self.assertIs(obj, i.rgen)
        # Check that the states are the same:
        state0 = rgens[0].get_state()['state']
        for i in rgens:
            self.assert_equal_npstates(state0, i.get_state()['state'])
        RandomGeneratorBorg.reset_state()

    def test_new_swarm(self):
        """Test that we make a new swarm without altering the old one."""
        rgenA = RandomGeneratorBorg(seed=0)
        rgb2 = RandomGeneratorBorg.make_new_swarm()
        rgenB = rgb2(seed=0)
        rgenC = RandomGeneratorBorg(seed=0)
        rgenD = rgb2(seed=0)

        assert rgenA.rgen is not rgenB.rgen  # Test new swarm
        assert rgenA.rgen is rgenC.rgen  # Test that the old swarm still works
        assert rgenB.rgen is rgenD.rgen  # Test that the new swarm works


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This package defines analysis tools for the PyRETIS program.

The analysis tools are intended to be used for analysis of the
simulation output from the PyRETIS program. The typical use of this
package is in post-processing of the results from a simulation (or
several simulations).

Package structure
-----------------

Modules
~~~~~~~

__init__.py
    This file, imports from the other modules. The method to analyse
    results from MD flux simulations is defined here since it will
    make use of analysis tools from `energy_analysis.py` and
    `order_analysis.py`.

analysis.py (:py:mod:`pyretis.analysis.analysis`)
    General methods for numerical analysis.

energy_analysis.py (:py:mod:`pyretis.analysis.energy_analysis`)
    Defines methods useful for analysing the energy output.

histogram.py (:py:mod:`pyretis.analysis.histogram`)
    Defines methods useful for generating histograms.

order_analysis.py (:py:mod:`pyretis.analysis.order_analysis`)
    Defines methods useful for analysis of order parameters.

path_analysis.py (:py:mod:`pyretis.analysis.path_analysis`)
    Defines methods for analysis of path ensembles.

Important methods defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

analyse_energies (:py:func:`.analyse_energies`)
    Analyse energy data from a simulation. It will calculate a running
    average, a distribution and do a block error analysis.

analyse_flux (:py:func:`.analyse_flux`)
    Analyse flux data from a MD flux simulation. It will calculate a
    running average, a distribution and do a block error analysis.

analyse_orderp (:py:func:`.analyse_orderp`)
    Analyse order parameter data. It will calculate a running average,
    a distribution and do a block error analysis. In addition, it will
    analyse the mean square displacement (if requested).

analyse_path_ensemble (:py:func:`.analyse_path_ensemble`)
    Analyse the results from a single path ensemble. It will calculate
    a running average of the probabilities, a crossing probability,
    perform a block error analysis, analyse lengths of paths,
    type of Monte Carlo moves and calculate an efficiency.

match_probabilities (:py:func:`.match_probabilities`)
    Method to match probabilities from several path simulations.
    Useful for obtaining the overall crossing probability.

histogram (:py:func:`.histogram`)
    Generates histogram, basically a wrapper around
    numpy's ``histogram``.

match_all_histograms (:py:func:`.match_all_histograms`)
    Method to match histograms from umbrella simulations.

retis_flux (:py:func:`.retis_flux`)
    Method for calculating the initial flux for RETIS simulations.

retis_rate (:py:func:`.retis_rate`)
    Method for calculating the rate constant for RETIS simulations.
"""
from .analysis import running_average, block_error, block_error_corr
from .energy_analysis import analyse_energies
from .flux_analysis import analyse_flux
from .histogram import histogram, match_all_histograms
from .order_analysis import analyse_orderp
from .path_analysis import (analyse_path_ensemble, analyse_repptis_ensemble,
                            match_probabilities, retis_flux, retis_rate,
                            perm_calculations)


def analyse_md_flux(crossdata, energydata, orderdata, settings):
    """Analyse the output from a MD-flux simulation.

    The obtained results will be returned as a convenient structure for
    plotting or reporting.

    Parameters
    ----------
    crossdata : numpy.array
        This is the data containing information about crossings.
    energydata : numpy.array
        This is the raw data for the energies.
    orderdata : numpy.array
        This is the raw data for the order parameter.
    settings : dict
        The settings for the analysis (e.g block length for error
        analysis) and some settings from the simulation (interfaces,
        time step etc.).

    Returns
    -------
    results : dict
        This dict contains the results from the different analysis as a
        dictionary. This dict can be used further for plotting or for
        generating reports.

    """
    results = {'flux': analyse_flux(crossdata, settings),
               'energy': analyse_energies(energydata, settings),
               'order': analyse_orderp(orderdata, settings)}
    return results



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the UmbrellaWindowSimulation class."""
import logging
import unittest
import numpy as np
from pyretis.core.units import create_conversion_factors
from pyretis.core.box import create_box
from pyretis.core.system import System
from pyretis.core.random_gen import RandomGenerator
from pyretis.core.particles import Particles
from pyretis.forcefield import ForceField
from pyretis.forcefield.potentials import DoubleWell, RectangularWell
from pyretis.simulation.mc_simulation import UmbrellaWindowSimulation
logging.disable(logging.CRITICAL)


def create_test_system():
    """Just set up and create a simple test system."""
    create_conversion_factors('reduced')
    box = create_box(periodic=[False])
    system = System(temperature=500, units='reduced', box=box)
    system.particles = Particles(dim=system.get_dim())
    system.add_particle(name='X', pos=np.array([-0.7]))
    potential_dw = DoubleWell(a=1, b=1, c=0.02)
    potential_rw = RectangularWell()
    system.forcefield = ForceField(
        'Double well with rectangular bias',
        potential=[potential_dw, potential_rw],
        params=[{'a': 1.0, 'b': 1.0, 'c': 0.02}, None]
    )
    return system


class TestUmbrellaWindow(unittest.TestCase):
    """Run the tests for UmbrellaWindow Simulation class."""

    def test_umbrellawindow(self):
        """Test that we can create the simulation object."""
        system = create_test_system()
        rgen = RandomGenerator()
        potential_rw = system.forcefield.potential[1]
        potential_rw.set_parameters({'left': -1.0, 'right': -0.4})
        system.potential()
        ensemble = {'rgen': rgen, 'system': system}
        settings = {'simulation': {'rgen': 'rgen',
                                   'umbrella': [-1.0, -0.4],
                                   'overlap': -0.5,
                                   'maxdx': 0.1}}
        controls = {'mincycle': 10}
        simulation = UmbrellaWindowSimulation(ensemble, settings, controls)
        for _ in simulation.run():
            pass
        self.assertTrue(system.particles.pos[0][0] > -0.5)
        # Test restart features:
        restart = simulation.restart_info()
        self.assertEqual(restart['type'], simulation.simulation_type)
        settings['restart'] = restart
        settings['restart']['overlap'] = 'NonSense'
        simulation.cycle['step'] = 999
        simulation = UmbrellaWindowSimulation(ensemble, settings, controls)
        self.assertEqual(simulation.overlap, 'NonSense')
        self.assertEqual(simulation.cycle['step'], 17)
        # Test the creation when the random generator is not given:
        ensemble = {'system': system}
        simulation = UmbrellaWindowSimulation(
            ensemble,
            settings={'simulation': {'umbrella': [-1.0, -0.4],
                                     'overlap': -0.5,
                                     'maxdx': 0.1}},
            controls={})
        self.assertIsInstance(simulation.rgen, type(rgen))

        with self.assertRaises(TypeError):
            simulation = UmbrellaWindowSimulation(ensemble,
                                                  settings=None,
                                                  controls={})


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""
Compare a PyRETIS simulation to LAMMPS.

In this example, we re-run a LAMMPS simulation using PyRETIS, and
compare different thermodynamic output.

"""
import sys
import tempfile
import colorama
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import gridspec
from pyretis.core.units import units_from_settings
from pyretis.setup import create_simulation
from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file


LAMMPS_HEAD = '#Step Temp Press PotEng KinEng TotEng Pxx Pyy Pzz Pxy Pxz Pyz'
LAMMPS_FMT = '{:8d}' + ' {:>12.7f}' * 11
TOL_ENERGY = 1.0
TOL_PRESS = 2.0


def set_up(settings_file):
    """Set up the simulation from a settings file."""
    print_to_screen(f'Loading settings: {settings_file}',
                    level='info')
    settings = parse_settings_file(settings_file)
    units_from_settings(settings)
    simulation = create_simulation(settings)
    return simulation


def format_thermo(result):
    """Format the thermo output similar to LAMMPS."""
    step = result['cycle']['step']
    thermo = result['thermo']
    press = thermo['press-tens']
    return LAMMPS_FMT.format(step, thermo['temp'], thermo['press'],
                             thermo['vpot'], thermo['ekin'], thermo['etot'],
                             press[0, 0], press[1, 1], press[2, 2],
                             press[0, 1], press[0, 2], press[1, 2])


def run_simulation(simulation, outputfile):
    """Run the simulation and write output to screen/file."""
    print_to_screen(f'Running simulation: {simulation}')
    with open(outputfile, 'w') as output:
        output.write(f'{LAMMPS_HEAD}\n')
        print(LAMMPS_HEAD)
        try:
            for step in simulation.run():
                txt = format_thermo(step)
                print(txt)
                output.write(f'{txt}\n')
        except KeyboardInterrupt:
            print_to_screen('Aborting simulation!')
            return 1
    return 0


def mean_squared_error(value1, value2):
    """Compute the mean squared error."""
    nmax = min(len(value1), len(value2))
    return ((value1[:nmax] - value2[:nmax])**2).mean()


def mean_absolute_percentage_error(value1, value2):
    """Compute the mean absolute percentage error."""
    nmax = min(len(value1), len(value2))
    return 100.0 * (np.abs(value1[:nmax] - value2[:nmax])).mean()


def plot_energy(lammps, pyret):
    """Just plot the energy terms."""
    print_to_screen('Plotting energy terms...')
    fig = plt.figure()
    grid = gridspec.GridSpec(3, 2)

    axes = (fig.add_subplot(grid[0, 0]),
            fig.add_subplot(grid[0, 1]),
            fig.add_subplot(grid[1, 0]),
            fig.add_subplot(grid[1, 1]),
            fig.add_subplot(grid[2, :]))

    labels = ('Temperature', 'Pressure', 'Potential energy',
              'Kinetic energy', 'Total energy')

    last = len(axes) - 1

    retval = 0

    for i, (axi, labi) in enumerate(zip(axes, labels)):
        linel, = axi.plot(lammps[:, 0], lammps[:, i + 1], lw=5, ls='-')
        linep, = axi.plot(pyret[:, 0], pyret[:, i + 1], lw=2, ls=':')
        axi.set_ylabel(labi)
        axi.set_xlabel('Step')
        mse = mean_squared_error(lammps[:, i + 1], pyret[:, i + 1])
        mape = mean_absolute_percentage_error(lammps[:, i + 1],
                                              pyret[:, i + 1])
        if mape > TOL_ENERGY:
            retval += 1
            print_to_screen(
                f'Too large MAPE for {labi}: {mape:.3e}',
                level='error'
            )
        rec = plt.Rectangle((0, 0), 1, 1, fill=False, edgecolor='none',
                            visible=False)
        if i < last:
            axi.legend(
                (rec, rec),
                (f'MSE = {mse:.3e}', f'MAPE = {mape:.3e}'),
                frameon=False
            )
        else:
            axi.legend((linel, linep, rec, rec),
                       ('LAMMPS', 'PyRETIS', f'MSE = {mse:.3e}',
                        f'MAPE = {mape:.3e}',),
                       frameon=False, ncol=2, columnspacing=0.0)
    fig.tight_layout()
    return retval


def plot_pressure(lammps, pyret):
    """Plot the pressure components."""
    print_to_screen('Plotting pressure terms...')
    fig = plt.figure()
    grid = gridspec.GridSpec(3, 2)
    last = 5
    retval = 0
    for i, term in enumerate(('Pxx', 'Pyy', 'Pzz', 'Pxy', 'Pxz', 'Pyz')):
        axi = fig.add_subplot(grid[i % 3, int(i / 3)])
        linel, = axi.plot(lammps[:, 0], lammps[:, i + 6], lw=5, ls='-')
        linep, = axi.plot(pyret[:, 0], pyret[:, i + 6], lw=2, ls=':')
        axi.set_ylabel(term)
        mse = mean_squared_error(lammps[:, i + 6], pyret[:, i + 6])
        mape = mean_absolute_percentage_error(lammps[:, i + 6],
                                              pyret[:, i + 6])
        if mape > TOL_PRESS:
            retval += 1
            print_to_screen(
                f'Too large MAPE for {term}: {mape:.3e}',
                level='error'
            )
        rec = plt.Rectangle((0, 0), 1, 1, fill=False, edgecolor='none',
                            visible=False)
        if i < last:
            axi.legend(
                (rec, rec),
                (f'MSE = {mse:.3e}', f'MAPE = {mape:.3e}'),
                frameon=False
            )
        else:
            axi.legend((linel, linep, rec, rec),
                       ('LAMMPS', 'PyRETIS',
                        f'MSE = {mse:.3e}',
                        f'MAPE = {mape:.3e}'),
                       frameon=False, ncol=1)
    fig.tight_layout()
    return retval


def plot_all_vs(lammps, pyret):
    """Plot terms vs each other."""
    print_to_screen('Plotting all terms...')
    fig = plt.figure()
    nrow, ncol = 3, 4
    grid = gridspec.GridSpec(nrow, ncol)
    labels = ('Temperature', 'Pressure', 'Potential energy',
              'Kinetic energy', 'Total energy', 'Pxx', 'Pyy',
              'Pzz', 'Pxy', 'Pxz', 'Pyz')
    nmax = min(len(lammps), len(pyret))
    for i, labi in enumerate(labels):
        row = i // ncol
        col = i % ncol
        axi = fig.add_subplot(grid[row, col])
        xval = lammps[:nmax, i + 1]
        yval = pyret[:nmax, i + 1]
        axi.scatter(xval, yval)
        axi.set_xlabel('LAMMPS')
        axi.set_ylabel('PyRETIS')
        axi.set_title(labi)
        fit_line = np.poly1d(np.polyfit(xval, yval, 1))
        newx = np.linspace(min(xval), max(xval), 3)
        rval = np.corrcoef(fit_line(xval), yval)
        # Since scatter don't influence the color cycler:
        axi.plot([], [], lw=3)
        linef, = axi.plot(newx, fit_line(newx), lw=3)
        axi.legend((linef, ), (f'Rsq = {rval[0, 1] ** 2:.3f}', ),
                   frameon=False)
    fig.tight_layout()


def plot_comparison(lammps_file, pyretis_file):
    """Plot comparison with LAMMPS."""
    print_to_screen(f'Loading LAMMPS data: {lammps_file}',
                    level='info')
    lammps = np.loadtxt(lammps_file)
    print_to_screen(f'Loading PyRETIS data: {pyretis_file}',
                    level='info')
    pyret = np.loadtxt(pyretis_file)
    if lammps.shape != pyret.shape:
        print_to_screen('LAMMPS and PyRETIS data have different shape!',
                        level='warning')

    plt.style.use('seaborn-v0_8-poster')
    ret1 = plot_energy(lammps, pyret)
    if ret1 == 0:
        print_to_screen('RMSD comparison for thermodynamic properties is OK!',
                        level='info')
    ret2 = plot_pressure(lammps, pyret)
    if ret2 == 0:
        print_to_screen('RMSD comparison for pressure components is OK!',
                        level='info')
    # plot_all_vs(lammps, pyret)

    if len(pyret) < len(lammps):
        print_to_screen(
            f'PyRETIS data too short: {len(pyret)} vs LAMMPS: {len(lammps)}',
            level='error'
        )
        return 1
    return ret1 + ret2


def main(settings_file, lammps_file, show_plot=True):
    """Parse input, run simulation and the comparison."""
    simulation = set_up(settings_file)
    retval = 1
    with tempfile.NamedTemporaryFile(prefix='thermo_', suffix='.txt') as tmp:
        run_simulation(simulation, tmp.name)
        retval = plot_comparison(lammps_file, tmp.name)
        if show_plot:
            plt.show()
        if retval == 0:
            print_to_screen('Comparison was successful!', level='success')
        else:
            print_to_screen('Comparison FAILED!', level='error')
    return retval


if __name__ == '__main__':
    colorama.init(autoreset=True)
    sys.exit(main(sys.argv[1], sys.argv[2], show_plot='plot' in sys.argv[1:]))



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Plot results using matplotlib."""
import numpy as np
from matplotlib import pyplot as plt


def make_plot():
    """Create the plot and show it."""
    plt.style.use('seaborn-poster')
    raw_data = np.loadtxt('thermo.txt')
    data = {}
    for i, key in enumerate(('step', 'temp', 'pot', 'kin', 'tot', 'press')):
        data[key] = raw_data[:, i]

    fig = plt.figure()
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)
    ax1.grid()
    ax2.grid()
    ax1.plot(data['step'], data['pot'], lw=3, alpha=0.8,
             label='Potential')
    ax1.plot(data['step'], data['kin'], lw=3,
             alpha=0.8, label='Kinetic')
    ax1.plot(data['step'], data['tot'], lw=3,
             alpha=0.8, label='Total')
    ax1.set_xlabel('Step')
    ax1.set_ylabel('Energy per particle')
    ax1.legend()
    ax1.set_xlim(data['step'][0], data['step'][-1])
    linep, = ax2.plot(data['step'], data['press'], lw=3,
                      alpha=0.8, label='Pressure')
    ax3 = ax2.twinx()

    color_cycle = ax2._get_lines.prop_cycler
    color = next(color_cycle)['color']
    linet, = ax3.plot(data['step'], data['temp'], lw=3,
                      alpha=0.8, label='Temperature', color=color)
    ax2.set_ylabel('Pressure')
    ax3.set_ylabel('Temperature')
    ax2.legend([linep, linet], ['Pressure', 'Temperature'], loc='center')
    ax2.set_xlim(data['step'][0], data['step'][-1])
    fig.tight_layout()
    plt.show()


if __name__ == '__main__':
    make_plot()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module defines a class, representing a collection of particles.

The class for particles is, in reality, a simplistic particle list which
stores positions, velocities, masses etc. and is used for representing
the particles in the simulations.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Particles (:py:class:`.Particles`)
    Class for a list of particles.

ParticlesExt (:py:class:`.ParticlesExt`)
    Class for an external particle list.

"""
from copy import copy
import logging
import numpy as np
from pyretis.core.common import compare_objects
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = ['Particles', 'ParticlesExt']


class Particles:
    """Base class for a collection of particles.

    This is a simple particle list. It stores the positions,
    velocities, forces, masses (and inverse masses) and type
    information for a set of particles. This class also defines a
    method for iterating over pairs of particles, which could be
    useful for implementing neighbour lists. In this particular
    class, this method will just define an all-pairs list.

    Attributes
    ----------
    npart : integer
        Number of particles.
    pos : numpy.array
        Positions of the particles.
    vel : numpy.array
        Velocities of the particles.
    force : numpy.array
        Forces on the particles.
    virial : numpy.array
        The current virial for the particles.
    mass : numpy.array
        Masses of the particles.
    imass : numpy.array
        Inverse masses, `1.0 / self.mass`.
    name : list of strings
        A name for the particle. This may be used as short text
        describing the particle.
    ptype : numpy.array of integers
        A type for the particle. Particles with identical `ptype` are
        of the same kind.
    dim : int
        This variable is the dimensionality of the particle list. This
        should be derived from the box. For some functions, it is
        convenient to be able to access the dimensionality directly
        from the particle list. It is therefore set as an attribute
        here.
    vpot : float
        The potential energy of the particles.
    ekin : float
        The kinetic energy of the particles.

    """

    particle_type = 'internal'

    # Attributes to store when restarting/copying:
    _copy_attr = {'npart', 'name', 'ptype', 'dim'}
    # Attributes which are numpy arrays:
    _numpy_attr = {'pos', 'vel', 'force', 'virial', 'mass', 'imass',
                   'ptype', 'ekin', 'vpot'}

    def __init__(self, dim=1):
        """Initialise the Particle list.

        Here we just create an empty particle list.

        Parameters
        ----------
        dim : integer, optional
            The number of dimensions we are considering for positions,
            velocities and forces.

        """
        self.npart = 0
        self.pos = None
        self.vel = None
        self.vpot = None
        self.ekin = None
        self.force = None
        self.mass = None
        self.imass = None
        self.name = []
        self.ptype = None
        self.virial = None
        self.dim = dim

    def empty_list(self):
        """Reset the particle list.

        This will delete all particles in the list and set other
        variables to `None`.

        Note
        ----
        This is almost `self.__init__` repeated. The reason for this is
        simply that we want to define all attributes in `self.__init__`
        and not get any 'surprise attributes' defined elsewhere.
        Also, note that the dimensionality (`self.dim`) is not changed
        in this method.

        """
        self.npart = 0
        self.pos = None
        self.vpot = None
        self.ekin = None
        self.vel = None
        self.force = None
        self.mass = None
        self.imass = None
        self.name = []
        self.ptype = None
        self.virial = None

    def _copy_attribute(self, attr, copy_function):
        """Copy an attribute.

        Parameters
        ----------
        attr : string
            The attribute to copy.
        copy_function : callable
            The method to use for copying the attribute.

        Returns
        -------
        out : object
            A copy of the selected attribute.

        """
        val = getattr(self, attr, None)
        if val is None:
            return None
        return copy_function(val)

    def copy(self):
        """Return a copy of the particle state.

        Returns
        -------
        out : object like :py:class:`.Particles`
            A copy of the current state of the particles.

        """
        particles_copy = self.__class__(dim=self.dim)
        for attr in self._copy_attr:
            copy_attr = self._copy_attribute(attr, copy)
            setattr(particles_copy, attr, copy_attr)
        for attr in self._numpy_attr:
            copy_attr = self._copy_attribute(attr, np.copy)
            setattr(particles_copy, attr, copy_attr)
        return particles_copy

    def __eq__(self, other):
        """Compare two particle states."""
        attrs = self._copy_attr.union(self._numpy_attr)
        return compare_objects(self, other, attrs,
                               numpy_attrs=self._numpy_attr)

    def set_pos(self, pos):
        """Set positions for the particles.

        Parameters
        ----------
        pos : numpy.array
            The positions to set.

        """
        self.pos = np.copy(pos)

    def get_pos(self):
        """Return (a copy of) positions."""
        return np.copy(self.pos)

    def set_vel(self, vel):
        """Set velocities for the particles.

        Parameters
        ----------
        vel : numpy.array
            The velocities to set.

        """
        self.vel = np.copy(vel)

    def get_vel(self):
        """Return (a copy of) the velocities."""
        return np.copy(self.vel)

    def set_force(self, force):
        """Set the forces for the particles.

        Parameters
        ----------
        force : numpy.array
            The forces to set.

        """
        self.force = np.copy(force)

    def get_force(self):
        """Return (a copy of) the forces."""
        return np.copy(self.force)

    def add_particle(self, pos, vel, force, mass=1.0,
                     name='?', ptype=0):
        """Add a particle to the system.

        Parameters
        ----------
        pos : numpy.array
            Positions of new particle.
        vel :  numpy.array
            Velocities of new particle.
        force : numpy.array
            Forces on the new particle.
        mass : float, optional
            The mass of the particle.
        name : string, optional
            The name of the particle.
        ptype : integer, optional
            The particle type.

        """
        if self.npart == 0:
            self.name = [name]
            self.ptype = np.array(ptype, dtype='i4')
            self.pos = np.zeros((1, self.dim))
            self.pos[0] = pos
            self.vel = np.zeros((1, self.dim))
            self.vel[0] = vel
            self.force = np.zeros((1, self.dim))
            self.force[0] = force
            self.mass = np.zeros((1, 1))  # Column matrix.
            self.mass[0] = mass
            self.imass = 1.0 / self.mass
        else:
            self.name.append(name)
            self.ptype = np.append(self.ptype, ptype)
            self.pos = np.vstack([self.pos, pos])
            self.vel = np.vstack([self.vel, vel])
            self.force = np.vstack([self.force, force])
            self.mass = np.vstack([self.mass, mass])
            self.imass = np.vstack([self.imass, 1.0/mass])
        self.npart += 1

    def get_selection(self, properties, selection=None):
        """Return selected properties for a selection of particles.

        Parameters
        ----------
        properties : list of strings
            The strings represent the properties to return.
        selection : list with indices to return, optional
            If a selection is not given, data for all particles
            are returned.

        Returns
        -------
        out : list
            A list with the properties in the order they were asked for
            in the ``properties`` argument.

        """
        sel_prop = []
        for prop in properties:
            if hasattr(self, prop):
                var = getattr(self, prop)
                if isinstance(var, list):
                    if selection is None:
                        sel_prop.append(var)
                    else:
                        sel_prop.append([var[i] for i in selection])
                else:
                    if selection is None:
                        sel_prop.append(var)
                    else:
                        sel_prop.append(var[selection])
        return sel_prop

    def __iter__(self):
        """Iterate over the particles.

        This function will yield the properties of the different
        particles.

        Yields
        ------
        out : dict
            The information in `self.pos`, `self.vel`, ... etc.

        """
        for i, pos in enumerate(self.pos):
            part = {'pos': pos, 'vel': self.vel[i], 'force': self.force[i],
                    'mass': self.mass[i], 'imass': self.imass[i],
                    'name': self.name[i], 'ptype': self.ptype[i]}
            yield part

    def pairs(self):
        """Iterate over all pairs of particles.

        Yields
        ------
        out[0] : integer
            The index for the first particle in the pair.
        out[1] : integer
            The index for the second particle in the pair.
        out[2] : integer
            The particle type of the first particle.
        out[3] : integer
            The particle type of the second particle.

        """
        for i, itype in enumerate(self.ptype[:-1]):
            for j, jtype in enumerate(self.ptype[i+1:]):
                yield (i, i+1+j, itype, jtype)

    def __str__(self):
        """Print out basic info about the particle list."""
        basic_info = f'Particles: {self.npart}\n' + \
                     f'Types: {np.unique(self.ptype)}\n' + \
                     f'Names: {set(self.name)}'
        return basic_info

    def restart_info(self):
        """Generate information for saving a restart file."""
        info = {'type': self.particle_type}
        for copy_list in (self._copy_attr, self._numpy_attr):
            for attr in copy_list:
                try:
                    info[attr] = getattr(self, attr)
                except AttributeError:
                    logger.warning(('Missing attribute "%s" when creating'
                                    ' restart information.'), attr)
                    info[attr] = None
        return info

    def load_restart_info(self, info):
        """Load restart information.

        Parameters
        ----------
        info : dict
            Dictionary with the settings to load.

        """
        for attr in self._copy_attr.union(self._numpy_attr):
            if attr in info:
                setattr(self, attr, info[attr])
            else:
                msg = f'Could not set "{attr}" for particles ' + \
                       'from restart info'
                logger.error(msg)
                raise ValueError(msg)

    def reverse_velocities(self):
        """Reverse the velocities in the system."""
        self.vel = self.vel * -1


class ParticlesExt(Particles):
    """A particle list, when positions and velocities are stored in files.

    Attributes
    ----------
    config : tuple
        The file name and index in this file for the configuration
        the particle list is representing.
    vel_rev : boolean
        If this is True, the velocities in the file represeting
        the configuration will have to be reversed before they are
        used.

    """

    particle_type = 'external'

    # Attributes to store when restarting/copying:
    _copy_attr = {'npart', 'name', 'ptype', 'dim',
                  'config', 'vel_rev'}
    # Attributes which are numpy arrays:
    _numpy_attr = {'pos', 'vel', 'force', 'virial', 'mass', 'imass',
                   'ptype', 'ekin', 'vpot'}

    def __init__(self, dim=1):
        """Create an empty ParticleExt list.

        Parameters
        ----------
        dim : integer, optional
            The number of dimensions we are considering for positions,
            velocities and forces.

        """
        super().__init__(dim=dim)
        self.config = (None, None)
        self.vel_rev = False

    def add_particle(self, pos, vel, force, mass=1.0,
                     name='?', ptype=0):
        """Add a particle to the system.

        Parameters
        ----------
        pos : tuple
            Positions of new particle.
        vel : boolean
            Velocities of new particle.
        force : tuple
            Forces on the new particle.
        mass : float, optional
            The mass of the particle.
        name : string, optional
            The name of the particle.
        ptype : integer, optional
            The particle type.

        """
        self.name = [name]
        self.ptype = np.array(ptype, dtype='i4')
        self.pos = None
        self.set_pos(pos)
        self.vel = None
        self.set_vel(vel)
        self.force = force
        self.mass = np.zeros((1, 1))  # Column matrix.
        self.mass[0] = mass
        self.imass = 1.0 / self.mass
        self.npart = 1

    def empty_list(self):
        """Just empty the list."""
        super().empty_list()
        self.config = (None, None)
        self.vel_rev = False

    def reverse_velocities(self):
        """Reverse the velocities in the system."""
        self.vel_rev = not self.vel_rev

    def set_pos(self, pos):
        """Set positions for the particles.

        This will copy the input positions, for this class, the
        input positions are assumed to be a file name with a
        corresponding integer which determines the index for the
        positions in the file for cases where the file contains
        several snapshots.

        Parameters
        ----------
        pos : tuple of (string, int)
            The positions to set, this represents the file name and the
            index for the frame in this file.

        """
        self.config = (pos[0], pos[1])

    def get_pos(self):
        """Just return the positions of the particles."""
        return self.config

    def set_vel(self, rev_vel):
        """Set velocities for the particles.

        Here we store information which tells if the
        velocities should be reversed or not.

        Parameters
        ----------
        rev_vel : boolean
            The velocities to set. If True, the velocities should
            be reversed before used.

        """
        self.vel_rev = rev_vel

    def get_vel(self):
        """Return info about the velocities."""
        return self.vel_rev

    def __str__(self):
        """Print out basic info about the particle list."""
        return f'Config: {self.config}\nReverse velocities: {self.vel_rev}'


def get_particle_type(engine_type):
    """Return the particle class consistent with a given engine.

    Parameters
    ----------
    engine_type : string
        The type of particles we are requesting.

    """
    particle_map = {'internal': Particles,
                    'external': ParticlesExt}
    try:
        return particle_map[engine_type]
    except KeyError:
        msg = f'Unknown particle type "{engine_type}" requested.'
        logger.critical(msg)
        raise ValueError(msg)


def particles_from_restart(restart):
    """Create particles from restart information.

    Parameters
    ----------
    restart : dict
        The restart settings.

    Returns
    -------
    particles : object like :py:class:`.Particles.`
        The object created from the restart information.

    """
    restart_particles = restart.get('particles', None)
    if restart_particles is None:
        logger.info('No particles were created from restart information.')
        return None
    klass = get_particle_type(restart_particles['type'])
    particles = klass(dim=restart_particles['dim'])
    particles.load_restart_info(restart_particles)
    return particles



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This sub-package handles data formats for PyRETIS.

The methods and classes defined in this package will format the data
created by PyRETIS and also enable the reading of such output data.

Further, this package includes some methods for interacting with other
simulation packages.

Package structure
-----------------

Modules
~~~~~~~

formatter.py (:py:mod:`pyretis.inout.formats.formatter`)
    Generic methods for formatting output from PyRETIS. Defines
    the base class for all formatters.

cp2k.py (:py:mod:`pyretis.inout.formats.cp2k`)
    A module defining input/output methods for use with CP2k.

cross.py (:py:mod:`pyretis.inout.formats.cross`)
    Formatting of crossing data from PyRETIS.

energy.py (:py:mod:`pyretis.inout.formats.energy`)
    Formatting of energy data from PyRETIS.

gromacs.py (:py:mod:`pyretis.inout.formats.gromacs`)
    A module defining input/output methods for use with GROMACS.

__init__.py
    This file.

order.py (:py:mod:`pyretis.inout.formats.order`)
    Formatting of order parameter data from PyRETIS.

path.py (:py:mod:`pyretis.inout.formats.path`)
    Formatting of path/trajectory data from PyRETIS.

pathensemble.py (:py:mod:`pyretis.inout.formats.pathensemble`)
    Formatting of path ensemble data from PyRETIS.

snapshot.py (:py:mod:`pyretis.inout.formats.snapshot`)
    Formatting of snapshot data from PyRETIS.

txt_table.py (:py:mod:`pyretis.inout.formats.txt_table`)
    Definition of a table-like format.

xyz.py (:py:mod:`pyretis.inout.formats.xyz`)
    A module defining input/output methods for use with a XYZ format.


Important classes defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CrossFormatter (:py:class:`.CrossFormatter`)
    A class for formatting crossing data from flux simulations.

EnergyFormatter (:py:class:`.EnergyFormatter`)
    A class for formatting energy data from PyRETIS.

EnergyPathFormatter (:py:class:`.EnergyPathFormatter`)
    A class for formatting energy data for paths.

OrderFormatter (:py:class:`.OrderFormatter`)
    A class for formatting order parameter data.

OrderPathFormatter (:py:class:`.OrderPathFormatter`)
    A class for formatting order parameter data for paths.

PathExtFormatter (:py:class:`.PathExtFormatter`)
    A class for formatting external trajectories.

PathIntFormatter (:py:class:`.PathIntFormatter`)
    A class for formatting internal trajectories.

PathEnsembleFormatter (:py:class:`.PathEnsembleFormatter`)
    A class for formatting path ensemble data.

SnapshotFormatter (:py:class:`.SnapshotFormatter`)
    Generic class for formatting system snapshots (coordinates).

TxtTableFormatter (:py:class:`.TxtTableFormatter`)
    A class for creating a table-like format.

"""
import logging
from pyretis.inout.formats.cross import CrossFormatter, CrossFile
from pyretis.inout.formats.energy import (
    EnergyFormatter,
    EnergyPathFormatter,
    EnergyFile,
    EnergyPathFile,
)
from pyretis.inout.formats.order import (
    OrderFormatter,
    OrderPathFormatter,
    OrderFile,
    OrderPathFile,
)
from pyretis.inout.formats.path import (
    PathExtFormatter,
    PathIntFormatter,
    PathExtFile,
    PathIntFile,
)
from pyretis.inout.formats.pathensemble import (
    PathEnsembleFormatter,
    PathEnsembleFile,
)
from pyretis.inout.formats.snapshot import (
    SnapshotFormatter,
    SnapshotFile,
)
from pyretis.inout.formats.txt_table import (
    TxtTableFormatter,
    PathTableFormatter,
    ThermoTableFormatter,
    RETISResultFormatter,
)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())
